
from . import *
from clifford import NUMBA_PARALLEL, MVArray
import itertools
from .rotor_parameterisation import general_logarithm

imt_func = layout.imt_func
gmt_func = layout.gmt_func
adjoint_func = layout.adjoint_func
e4_val = e4.value
e5_val = e5.value
ninf_val = einf.value


sparse_cost_imt = layout.imt_func_generator(grades_a=[0, 2, 4], grades_b=[1])
sparse_cost_gmt = layout.gmt_func_generator(grades_a=[0, 2, 4], grades_b=[0, 2, 4])


def check_p(Q):
    """
    For leo dorsts check product
    """
    return Q(0, 1, 3) - Q(2, 4, 5)


def check_p_cost(P, Q):
    """
    For leo dorsts check product cost
    """
    return (P|check_p(Q))[0]


def point_to_line_cluster_distance(point, line_cluster):
    """
    Distance between a single point and a cluster of lines
    """
    return val_point_to_line_cluster_distance(point.value, np.array([l.value for l in line_cluster]))


@numba.njit
def val_point_to_line_cluster_distance(point_val, line_cluster_array):
    """
    Distance between a single point and a cluster of lines
    """
    error_val = 0.0
    for i in range(line_cluster_array.shape[0]):
        l_val = line_cluster_array[i, :]
        error_val -= imt_func(point_val, (gmt_func(gmt_func(l_val, point_val), l_val)))[0]
    return error_val


def midpoint_and_error_of_line_cluster_eig(line_cluster):
    """
    Gets an approximate center point of a line cluster
    as well as an estimate of the error
    Uses the eigenvalue and explicit derivatives
    """
    line_cluster_array = np.array([l.value for l in line_cluster], dtype=np.float64)
    mat2solve = val_truncated_get_line_reflection_matrix(line_cluster_array, 128)
    start = imt_func(no_val, sum(l.value for l in line_cluster))
    start = gmt_func(gmt_func(start, ninf_val), start)[1:6]

    point_val = np.zeros(32)
    point_val[1:6] = np.matmul(mat2solve, start)
    new_mv = layout.MultiVector(value=point_val)
    new_mv = normalise_n_minus_1((new_mv * einf * new_mv)(1))
    return new_mv, val_point_to_line_cluster_distance(new_mv.value, line_cluster_array)


def midpoint_and_error_of_line_cluster_svd(line_cluster):
    """
    Gets an approximate center point of a line cluster
    as well as an estimate of the error
    Uses the SVD and explicit derivatives
    """
    line_cluster_array = np.array([l.value for l in line_cluster])
    mat2solve = get_line_reflection_matrix(line_cluster)

    grade_val = 1
    column_mask = np.array(layout.gradeList) == grade_val
    mat_test = mat2solve[:, column_mask][1:6]
    print(mat_test)
    w, v = np.linalg.eig(mat_test)

    point_val = np.zeros(32)
    point_val[np.array(layout.gradeList) == grade_val] = v[:, 1]
    new_mv = layout.MultiVector(value=point_val)
    # new_mv = normalise_n_minus_1(new_mv * einf * new_mv)
    new_point = normalise_n_minus_1(new_mv)  # up(down(new_mv) / 2)
    return new_point, val_point_to_line_cluster_distance(new_point.value, line_cluster_array)


def midpoint_and_error_of_line_cluster(line_cluster):
    """
    Gets an approximate center point of a line cluster
    as well as an estimate of the error
    Hadfield and Lasenby AGACSE2018
    """
    line_cluster_array = np.array([l.value for l in line_cluster])
    cp_val = val_midpoint_of_line_cluster(line_cluster_array)
    return layout.MultiVector(value=cp_val), val_point_to_line_cluster_distance(cp_val, line_cluster_array)


def midpoint_and_error_of_line_cluster_grad(line_cluster):
    """
    Gets an approximate center point of a line cluster
    as well as an estimate of the error
    Hadfield and Lasenby AGACSE2018
    """
    line_cluster_array = np.array([l.value for l in line_cluster])
    cp_val = val_midpoint_of_line_cluster_grad(line_cluster_array)
    return layout.MultiVector(value=cp_val), val_point_to_line_cluster_distance(cp_val, line_cluster_array)


def line_plane_cost(line, plane):
    """
    A cost function for a line and a plane
    """
    P = normalised((line|plane)*I5)
    L = normalised(meet(P, plane))
    return line_cost_function(L, line)


def midline_and_error_of_plane_cluster(plane_cluster):
    """
    Finds the line that is the best intersection of the planes
    """
    plane_perms = itertools.permutations(plane_cluster, 2)
    line_list = [normalised(meet(pp[0], pp[1])(3)) for pp in plane_perms]

    ref_line = line_list[0]
    line_sum = 0.0 * e1
    for l in line_list:
        if (ref_line | l)[0] > 0:
            line_sum += l
        else:
            line_sum -= l
    line_average = average_objects([line_sum(3)])
    cost_val = 0.0
    for plane in plane_cluster:
        cost_val += line_plane_cost(line_average, plane)
    return line_average, cost_val


@numba.njit
def val_rotor_cost_sparse(R_val):
    """
    Evaluates Eivind Eiede's cost function of a rotor
    """
    rotation_val = R_val.copy()
    rotation_val[0] -= 1
    translation_val = sparse_cost_imt(R_val, e4_val)
    a = abs(float(sparse_cost_gmt(rotation_val, adjoint_func(rotation_val))[0]))
    b = abs(float(gmt_func(translation_val, adjoint_func(translation_val))[0]))
    return a + b


def rotor_cost(R):
    """
    Evaluates Eivind Eiede's cost function of a rotor
    """
    return val_rotor_cost_sparse(R.value)


gradeList = np.array(layout.gradeList)
grade_obj_func = cf.grade_obj_func
max_float = np.finfo(float).max
@numba.njit
def val_object_cost_function(obj_a_val, obj_b_val, motor=False):
    """
    Evaluates the rotor cost function between two objects
    """
    grade_a = grade_obj_func(obj_a_val, gradeList, 0.00001)
    grade_b = grade_obj_func(obj_b_val, gradeList, 0.00001)
    if grade_a != grade_b:
        return max_float
    else:
        if motor:
            R_val = val_motor_between_objects(obj_a_val, obj_b_val)
        else:
            R_val = val_rotor_between_objects_root(obj_a_val, obj_b_val)
        return np.abs(val_rotor_cost_sparse(R_val))


def alt_rotor_cost(V):
    """
    An alternate rotor cost for TRS rotors
    """
    logV = general_logarithm(V)
    scale_cost = np.abs(-2*logV[e45])
    scalefac = np.e**(-2*logV[e45])
    R = logV(e123)*e123
    rotation_cost = abs((R*~R)[0])
    translation_cost = scalefac*abs((logV - logV[e45]*e45 - logV(e123))|eo)
    return rotation_cost + scale_cost + translation_cost


def object_cost_function(obj_a, obj_b, motor=False, symmetric=False):
    """
    Evaluates the rotor cost function between two objects
    """
    if symmetric:
        return min([val_object_cost_function(obj_a.value, obj_b.value, motor=motor),
                    val_object_cost_function(obj_a.value, -obj_b.value, motor=motor)])
    return val_object_cost_function(obj_a.value, obj_b.value, motor=motor)


def object_set_log_cost_sum(object_set_a, object_set_b, object_type='generic'):
    """
    Evaluates the rotor cost function between two sets of objects
    """
    sum_val = 0.0
    if object_type == 'lines':
        for a, b in zip(object_set_a, object_set_b):
            sum_val += np.log1p(line_cost_function(a, b))
    else:
        for a, b in zip(object_set_a, object_set_b):
            sum_val += np.log1p(object_cost_function(a, b))
    return sum_val


def object_set_cost_sum(object_set_a, object_set_b,
                        object_type='generic', motor=False,
                        symmetric=False):
    """
    Evaluates the rotor cost function between two sets of objects
    """
    sum_val = 0.0
    if object_type == 'lines':
        for a, b in zip(object_set_a, object_set_b):
            sum_val += line_cost_function(a, b)
    else:
        for a, b in zip(object_set_a, object_set_b):
            sum_val += object_cost_function(a, b, motor=motor, symmetric=symmetric)
    return sum_val


@numba.njit(parallel=NUMBA_PARALLEL, nogil=True)
def val_object_set_cost_matrix(object_array_a, object_array_b):
    """
    Evaluates the rotor cost matrix between two sets of objects
    """
    matrix = np.zeros((object_array_a.shape[0], object_array_b.shape[0]))
    for a_ind in range(object_array_a.shape[0]):
        l_a = object_array_a[a_ind, :]
        for b_ind in range(object_array_b.shape[0]):
            l_b = object_array_b[b_ind, :]
            matrix[a_ind, b_ind] = val_object_cost_function(l_a, l_b)
    return matrix


@numba.njit(parallel=NUMBA_PARALLEL, nogil=True)
def val_line_set_cost_matrix(object_array_a, object_array_b):
    """
    Evaluates the rotor cost matrix between two sets of objects
    """
    matrix = np.zeros((object_array_a.shape[0], object_array_b.shape[0]))
    for a_ind in range(object_array_a.shape[0]):
        l_a = object_array_a[a_ind, :]
        for b_ind in range(object_array_b.shape[0]):
            l_b = object_array_b[b_ind, :]
            matrix[a_ind, b_ind] = val_line_cost_function(l_a, l_b)
    return matrix


def object_set_cost_matrix(object_set_a, object_set_b,
                           object_type="generic", symmetric=False):
    """
    Evaluates the rotor cost matrix between two sets of objects
    """
    object_array_a = MVArray(object_set_a).value
    object_array_b = MVArray(object_set_b).value
    if object_type == 'lines':
        ret_mat = val_line_set_cost_matrix(object_array_a, object_array_b)
        if symmetric:
            ret_mat = np.minimum(ret_mat, val_line_set_cost_matrix(object_array_a, -object_array_b))
        return ret_mat
    else:
        ret_mat = val_object_set_cost_matrix(object_array_a, object_array_b)
        if symmetric:
            ret_mat = np.minimum(ret_mat, val_object_set_cost_matrix(object_array_a, -object_array_b))
        return ret_mat


def object_set_cost_matrix_sum(object_set_a, object_set_b, object_type='generic'):
    """
    Evaluates the sum of all elements in the rotor cost matrix between two sets of objects
    """
    if (len(object_set_a) == 0) or (len(object_set_b) == 0):
        return np.inf
    return np.sum(object_set_cost_matrix(object_set_a, object_set_b, object_type))


@numba.njit
def val_line_cost_function(obj_a_val, obj_b_val):
    grade_a = grade_obj_func(obj_a_val, gradeList, 0.00001)
    grade_b = grade_obj_func(obj_b_val, gradeList, 0.00001)
    if grade_a != grade_b:
        return max_float
    else:
        if imt_func(obj_a_val, obj_b_val)[0] < 0:
            R_val = val_rotor_between_lines(obj_a_val, -obj_b_val)
        else:
            R_val = val_rotor_between_lines(obj_a_val, obj_b_val)
        return np.abs(val_rotor_cost_sparse(R_val))


def line_cost_function(obj_a, obj_b):
    """
    Evaluates the rotor cost function between two objects
    """
    return val_line_cost_function(obj_a.value, obj_b.value)



from functools import reduce
import operator

import pytest
import numpy as np
import numpy.testing

from clifford import Cl, randomMV, Frame, \
    conformalize, grade_obj, MultiVector, MVArray

import clifford


class TestClifford:

    @pytest.fixture(params=[3, 4, 5], ids='Cl({})'.format)
    def algebra(self, request):
        return Cl(request.param)

    def test_inverse(self, algebra):
        layout, blades = algebra
        a = 1. + blades['e1']
        with pytest.raises(ValueError):
            1 / a
        for i in range(10):
            a = randomMV(layout, grades=[0, 1])
            denominator = float(a(1)**2-a(0)**2)
            if abs(denominator) > 1.e-5:
                a_inv = (-a(0)/denominator) + ((1./denominator) * a(1))
                assert abs((a * a_inv)-1.) < 1.e-11
                assert abs((a_inv * a)-1.) < 1.e-11
                assert abs(a_inv - 1./a) < 1.e-11

    def test_pseudoscalar(self, algebra):
        """
        Check if the outer product of the basis elements
        is a linear multiple of the pseudoscalar
        """
        layout, blades = algebra
        ps = reduce(clifford.op, layout.blades_of_grade(1))
        ps2 = layout.pseudoScalar
        assert np.linalg.matrix_rank(np.column_stack((ps.value, ps2.value))) == 1

    def test_grade_masks(self, algebra):
        layout, blades = algebra
        A = layout.randomMV()
        for i in range(layout.dims + 1):
            np.testing.assert_almost_equal(A(i).value, A.value*layout.grade_mask(i))

    def test_rotor_mask(self, algebra):
        layout, blades = algebra
        rotor_m = layout.rotor_mask
        rotor_m_t = np.zeros(layout.gaDims)
        for _ in range(10):
            rotor_m_t += 100*np.abs(layout.randomRotor().value)
        np.testing.assert_almost_equal(rotor_m_t > 0, rotor_m)

    def test_exp(self):
        layout, blades = Cl(3)
        e12 = blades['e12']
        theta = np.linspace(0, 100 * np.pi, 101)
        a_list = [np.e**(t * e12) for t in theta]
        for a in a_list:
            np.testing.assert_almost_equal(abs(a), 1.0, 5)

    def test_exp_g4(self):
        '''
        a numerical test for the exponential of a bivector. truth was
        generated by results of clifford v0.82
        '''
        layout, blades = Cl(4)

        valB = np.array([
            -0.                 ,  0.                 ,  # noqa
             0.                 , -0.                 ,  # noqa
            -0.                 , -1.9546896043012914 ,  # noqa
             0.7069828848351363 , -0.22839793693302957,  # noqa
             1.0226966962560002 ,  1.8673816483342143 ,  # noqa
            -1.7694566455296474 , -0.                 ,  # noqa
            -0.                 ,  0.                 ,  # noqa
            -0.                 , -0.                    # noqa
        ])
        valexpB = np.array([
            -0.8154675764311629  ,  0.                  ,  # noqa
             0.                  ,  0.                  ,  # noqa
             0.                  ,  0.3393508714682218  ,  # noqa
             0.22959588097548828 , -0.1331099867581965  ,  # noqa
            -0.01536404898029994 ,  0.012688721722814184,  # noqa
             0.35678394795928464 ,  0.                  ,  # noqa
             0.                  ,  0.                  ,  # noqa
             0.                  , -0.14740840378445502    # noqa
        ])

        B = MultiVector(layout=layout, value=valB)
        expB = MultiVector(layout=layout, value=valexpB)
        np.testing.assert_almost_equal(np.exp(B)[0].value, expB.value)

    def test_inv_g4(self):
        '''
        a numerical test for the inverse of a MV. truth was
        generated by results of clifford v0.82
        '''
        layout, blades = Cl(4)
        valA = np.array([
            -0.3184271488037198 , -0.8751064635010213 ,  # noqa
            -1.5011710376191947 ,  1.7946332649746224 ,  # noqa
            -0.8899576254164621 , -0.3297631748225678 ,  # noqa
             0.04310366054166925,  1.3970365638677635 ,  # noqa
            -1.545423393858595  ,  1.7790215501876614 ,  # noqa
             0.4785341530609175 , -1.32279679741638   ,  # noqa
             0.5874769077573831 , -1.0227287710873676 ,  # noqa
             1.779673249468527  , -1.5415648119743852    # noqa
        ])

        valAinv = np.array([
             0.06673424072253006 , -0.005709960252678998,  # noqa
            -0.10758540037163118 ,  0.1805895938775471  ,  # noqa
             0.13919236400967427 ,  0.04123255613093294 ,  # noqa
            -0.015395162562329407, -0.1388977308136247  ,  # noqa
            -0.1462160646855434  , -0.1183453106997158  ,  # noqa
            -0.06961956152268277 ,  0.1396713851886765  ,  # noqa
            -0.02572904638749348 ,  0.02079613649197489 ,  # noqa
            -0.06933660606043765 , -0.05436077710009021    # noqa
        ])

        A = MultiVector(layout=layout, value=valA)
        Ainv = MultiVector(layout=layout, value=valAinv)

        np.testing.assert_almost_equal(A.inv().value, Ainv.value)

    def test_indexing(self):
        layout, blades = Cl(3)
        e12 = blades['e12']
        e1 = blades['e1']
        e2 = blades['e2']
        e3 = blades['e3']
        assert e12[e12] == 1
        assert e12[e3] == 0
        assert e12[(2, 1)] == -1

    def test_add_float64(self):
        '''
        test array_wrap method to take control addition from numpy array
        '''
        layout, blades = Cl(3)
        e1 = blades['e1']

        np.float64(1) + e1
        assert 1 + e1 == np.float64(1) + e1

        assert 1 + e1 == e1 + np.float64(1)

    def test_array_control(self):
        '''
        test methods to take control addition from numpy arrays
        '''
        layout, blades = Cl(3)
        e1 = blades['e1']
        e3 = blades['e3']
        e12 = blades['e12']

        for i in range(100):

            number_array = np.random.rand(4)

            output = e12+(e1*number_array)
            output2 = MVArray([e12+(e1*n) for n in number_array])
            np.testing.assert_almost_equal(output, output2)

            output = e12 + (e1 * number_array)
            output2 = MVArray([e12 + (e1 * n) for n in number_array])
            np.testing.assert_almost_equal(output, output2)

            output = (number_array*e1) + e12
            output2 = MVArray([(n*e1) + e12 for n in number_array])
            np.testing.assert_almost_equal(output, output2)

            output = number_array / e12
            output2 = MVArray([n / e12 for n in number_array])
            np.testing.assert_almost_equal(output, output2)

            output = (e1 / number_array)
            output2 = MVArray([(e1/n) for n in number_array])
            np.testing.assert_almost_equal(output, output2)

            output = ((e1 / number_array)*e3)/e12
            output2 = MVArray([((e1 / n)*e3)/e12 for n in number_array])
            np.testing.assert_almost_equal(output, output2)

    def test_array_overload(self, algebra):
        '''
        test overload operations
        '''
        layout, blades = algebra
        test_array = MVArray([layout.randomMV() for i in range(100)])

        normed_array = test_array.normal()
        other_array = np.array([t.normal().value for t in test_array])
        np.testing.assert_almost_equal(normed_array.value, other_array)

        dual_array = test_array.dual()
        other_array_2 = np.array([t.dual().value for t in test_array])
        np.testing.assert_almost_equal(dual_array.value, other_array_2)

    def test_comparison_operators(self):
        layout, blades = Cl(3)
        e1 = blades['e1']
        e2 = blades['e2']

        pytest.raises(TypeError, operator.lt, e1, e2)
        pytest.raises(TypeError, operator.le, e1, e2)
        pytest.raises(TypeError, operator.gt, e1, e2)
        pytest.raises(TypeError, operator.ge, e1, e2)

        assert operator.eq(e1, e1) is True
        assert operator.eq(e1, e2) is False
        assert operator.ne(e1, e1) is False
        assert operator.ne(e1, e2) is True

        assert operator.eq(e1, None) is False
        assert operator.ne(e1, None) is True

    def test_layout_comparison_operators(self):
        l3a, _ = Cl(3)
        l3b, _ = Cl(3)
        l4, _ = Cl(4)

        assert operator.eq(l3a, l3b) is True
        assert operator.eq(l3a, l4) is False
        assert operator.eq(l3a, None) is False

        assert operator.ne(l3a, l3b) is False
        assert operator.ne(l3a, l4) is True
        assert operator.ne(l3a, None) is True

    def test_mv_str(self):
        """ Test the __str__ magic method """
        layout, blades = Cl(3)
        e1 = blades['e1']
        e2 = blades['e2']
        e12 = blades['e12']

        assert str(e1) == "(1^e1)"
        assert str(1 + e1) == "1 + (1^e1)"
        assert str(-e1) == "-(1^e1)"
        assert str(1 - e1) == "1 - (1^e1)"

    @pytest.mark.parametrize('dtype', [np.int64, np.float32, np.float64])
    @pytest.mark.parametrize('func', [
        operator.add,
        operator.sub,
        operator.mul,
        operator.xor,  # outer product
        operator.or_,  # inner product
    ])
    def test_binary_op_preserves_dtype(self, dtype, func):
        """ test that simple binary ops on blades do not promote types """
        layout, blades = Cl(3)
        e1 = blades['e1'].astype(dtype)
        e2 = blades['e2'].astype(dtype)
        assert func(e1, np.int8(1)).value.dtype == dtype
        assert func(e1, e2).value.dtype == dtype

    @pytest.mark.parametrize('func', [
        operator.inv,
        operator.pos,
        operator.neg,
        MultiVector.gradeInvol,
    ])
    def test_unary_op_preserves_dtype(self, func):
        """ test that simple unary ops on blades do not promote types """
        layout, blades = Cl(3)
        e1 = blades['e1']
        assert func(e1).value.dtype == e1.value.dtype

    def test_indexing_blade_tuple(self):
        # gh-151
        layout, blades = Cl(3)
        mv = layout.MultiVector(value=np.arange(2**3) + 1)

        # one swap makes the sign flip
        assert mv[1, 2] == -mv[2, 1]
        assert mv[2, 3] == -mv[3, 2]
        assert mv[3, 1] == -mv[1, 3]

        assert mv[1, 2, 3] == -mv[2, 1, 3]
        assert mv[1, 2, 3] == -mv[1, 3, 2]

        # two swaps does not
        assert mv[1, 2, 3] == mv[2, 3, 1] == mv[3, 1, 2]
        assert mv[3, 2, 1] == mv[2, 1, 3] == mv[1, 3, 2]

        # three swaps does
        assert mv[1, 2, 3] == -mv[3, 2, 1]


class TestBasicConformal41:
    def test_metric(self):
        layout = Cl(4, 1)[0]
        e1 = layout.blades['e1']
        e2 = layout.blades['e2']
        e3 = layout.blades['e3']
        e4 = layout.blades['e4']
        e5 = layout.blades['e5']

        assert (e1 * e1)[0] == 1
        assert (e2 * e2)[0] == 1
        assert (e3 * e3)[0] == 1
        assert (e4 * e4)[0] == 1
        assert (e5 * e5)[0] == -1

    def test_factorise(self):
        layout_a = Cl(3)[0]
        layout, blades, stuff = conformalize(layout_a)
        e1 = layout.blades['e1']
        e2 = layout.blades['e2']
        e3 = layout.blades['e3']
        e4 = layout.blades['e4']
        e5 = layout.blades['e5']

        up = stuff['up']

        blade = up(e1 + 3*e2 + 4*e3)^up(5*e1 + 3.3*e2 + 10*e3)^up(-13.1*e1)

        basis, scale = blade.factorise()
        new_blade = (reduce(lambda a, b: a^b, basis)*scale)
        print(new_blade)
        print(blade)
        np.testing.assert_almost_equal(new_blade.value, blade.value, 5)

    def test_gp_op_ip(self):
        layout = Cl(4, 1)[0]
        e1 = layout.blades['e1']
        e2 = layout.blades['e2']
        e3 = layout.blades['e3']
        e4 = layout.blades['e4']
        e5 = layout.blades['e5']

        e123 = layout.blades['e123']
        np.testing.assert_almost_equal(e123.value, (e1 ^ e2 ^ e3).value)
        np.testing.assert_almost_equal(e123.value, (e1 * e2 * e3).value)

        e12345 = layout.blades['e12345']
        np.testing.assert_almost_equal(e12345.value, (e1 ^ e2 ^ e3 ^ e4 ^ e5).value)
        np.testing.assert_almost_equal(e12345.value, (e1 * e2 * e3 * e4 * e5).value)

        e12 = layout.blades['e12']
        np.testing.assert_almost_equal(-e12.value, (e2 ^ e1).value)

        t = np.zeros(32)
        t[0] = -1
        np.testing.assert_almost_equal(t, (e12*e12).value)

    def test_categorization(self):
        layout = Cl(3)[0]
        e1 = layout.blades['e1']
        e2 = layout.blades['e2']
        e3 = layout.blades['e3']

        blades = [
            layout.scalar,
            e1,
            e1 ^ e2,
            (e1 + e2) ^ e2,
        ]
        for b in blades:
            # all invertible blades are also versors
            assert b.isBlade()
            assert b.isVersor()

        versors = [
            1 + (e1^e2),
            e1 + (e1^e2^e3),
        ]
        for v in versors:
            assert not v.isBlade()
            assert v.isVersor()

        neither = [
            layout.scalar*0,
            1 + e1,
            1 + (e1^e2^e3)
        ]
        for n in neither:
            assert not n.isBlade()
            assert not n.isVersor()

    def test_blades_of_grade(self):
        layout = Cl(3)[0]
        e1 = layout.blades['e1']
        e2 = layout.blades['e2']
        e3 = layout.blades['e3']
        assert layout.blades_of_grade(0) == [layout.scalar]
        assert layout.blades_of_grade(1) == [e1, e2, e3]
        assert layout.blades_of_grade(2) == [e1^e2, e1^e3, e2^e3]
        assert layout.blades_of_grade(3) == [e1^e2^e3]


class TestBasicSpaceTime:
    def test_initialise(self):

        # Dirac Algebra  `D`
        D, D_blades = Cl(1, 3, names='d', firstIdx=0)

        # Pauli Algebra  `P`
        P, P_blades = Cl(3, names='p')

        # put elements of each in namespace
        locals().update(D_blades)
        locals().update(P_blades)


class TestBasicAlgebra:

    def test_gp_op_ip(self):
        layout = Cl(3)[0]
        e1 = layout.blades['e1']
        e2 = layout.blades['e2']
        e3 = layout.blades['e3']

        print('outer product')
        e123 = layout.blades['e123']
        np.testing.assert_almost_equal(e123.value, (e1 ^ e2 ^ e3).value)
        np.testing.assert_almost_equal(e123.value, (e1 * e2 * e3).value)

        print('outer product ordering')
        e12 = layout.blades['e12']
        np.testing.assert_almost_equal(-e12.value, (e2 ^ e1).value)

        print('outer product zeros')
        np.testing.assert_almost_equal(0, (e1 ^ e1).value)
        np.testing.assert_almost_equal(0, (e2 ^ e2).value)
        np.testing.assert_almost_equal(0, (e3 ^ e3).value)

        print('scalar outer product')
        np.testing.assert_almost_equal(((1 + 0 * e1) ^ (1 + 0 * e1)).value, (1 + 0 * e1).value)

        print('scalar inner product')
        np.testing.assert_almost_equal(((1 + 0 * e1) | (1 + 0 * e1)).value, 0)

    @pytest.fixture(
        params=[Cl(i) for i in [3, 4]] + [conformalize(Cl(3)[0])],
        ids=['Cl(3)', 'Cl(4)', 'conformal Cl(3)']
    )
    def algebra(self, request):
        return request.param

    def test_grade_obj(self, algebra):
        layout = algebra[0]
        for i in range(len(layout.sig)+1):
            mv = layout.randomMV()(i)
            assert i == grade_obj(mv)

    def test_left_multiplication_matrix(self, algebra):
        layout = algebra[0]
        for i in range(1000):
            mv = layout.randomMV()
            mv2 = layout.randomMV()
            np.testing.assert_almost_equal(np.matmul(layout.get_left_gmt_matrix(mv), mv2.value), (mv*mv2).value)

    def test_right_multiplication_matrix(self, algebra):
        layout = algebra[0]
        for i in range(1000):
            a = layout.randomMV()
            b = layout.randomMV()
            b_right = layout.get_right_gmt_matrix(b)
            res = a*b
            res2 = layout.MultiVector(value=b_right@a.value)
            np.testing.assert_almost_equal(res.value, res2.value)


class TestFrame:

    def check_inv(self, A):
        Ainv = None
        for k in range(3):
            try:
                Ainv = A.inv
            except ValueError:
                pass
        if Ainv is None:
            return
        for m, a in enumerate(A):
            for n, b in enumerate(A.inv):
                if m == n:
                    assert(a | b == 1)
                else:
                    assert(a | b == 0)

    @pytest.mark.parametrize(('p', 'q'), [
        (2, 0), (3, 0), (4, 0)
    ])
    def test_frame_inv(self, p, q):
        layout, blades = Cl(p, q)
        A = Frame(layout.randomV(p + q))
        self.check_inv(A)

    @pytest.mark.parametrize(('p', 'q'), [
        (2, 0), (3, 0), (4, 0)
    ])
    def test_innermorphic(self, p, q):
        layout, blades = Cl(p, q)

        A = Frame(layout.randomV(p+q))
        R = layout.randomRotor()
        B = Frame([R*a*~R for a in A])
        assert A.is_innermorphic_to(B)



# https://clifford.readthedocs.io/en/latest/SpaceTimeAlgebra.html

from . import Cl, BladeMap

# Dirac Algebra  `D`
D, D_blades = Cl(1, 3, firstIdx=0, names='d')

# Pauli Algebra  `P`
P, P_blades = Cl(3, names='p')

# put elements of each in namespace
locals().update(D_blades)
locals().update(P_blades)

bm = BladeMap([(d01, p1),
               (d02, p2),
               (d03, p3),
               (d12, p12),
               (d23, p23),
               (d13, p13),
               (d0123, p123)])


def split(X):
    '''
    implements the spacetime split.
    '''
    return bm(X.odd*d0+X.even)



import re
from functools import reduce
from typing import List

import numpy as np
import numba
import sparse

# TODO: move some of these functions to this file if they're not useful anywhere
# else
import clifford as cf
from . import (
    get_adjoint_function,
    construct_tables,
    get_mult_function,
    get_leftLaInv,
    generate_blade_tup_map,
    generate_bitmap_to_linear_index_map,
    val_get_left_gmt_matrix,
    val_get_right_gmt_matrix,
)

from .io import read_ga_file


# The blade finding regex for parsing strings of mvs
_blade_pattern = re.compile(r"""
    ((^|\s)-?\s?\d+(\.\d+)?)\s|
    ((^|\+|-)\s?(\d+((e(\+|-))|\.)?(\d+)?)\^e\d+(\s|$))
""", re.VERBOSE)


class Layout(object):
    """ Layout stores information regarding the geometric algebra itself and the
    internal representation of multivectors.

    Parameters
    ----------

    signature : List[int]
        The signature of the vector space.  This should be
        a list of positive and negative numbers where the sign determines the
        sign of the inner product of the corresponding vector with itself.
        The values are irrelevant except for sign.  This list also determines
        the dimensionality of the vectors.  Signatures with zeroes are not
        permitted at this time.

        Examples:
          signature = [+1, -1, -1, -1] # Hestenes', et al. Space-Time Algebra
          signature = [+1, +1, +1]     # 3-D Euclidean signature

    bladeTupList : List[Tuple[int, ...]]
        List of tuples corresponding to the blades in the whole
        algebra.  This list determines the order of coefficients in the
        internal representation of multivectors.  The entry for the scalar
        must be an empty tuple, and the entries for grade-1 vectors must be
        singleton tuples.  Remember, the length of the list will be 2**dims.

        Example:
          bladeTupList = [(), (0,), (1,), (0, 1)]  # 2-D

    firstIdx : int
        The index of the first vector.  That is, some systems number
        the base vectors starting with 0, some with 1.  Choose by passing
        the correct number as firstIdx.  0 is the default.

    names : List[str]
        List of names of each blade.  When pretty-printing multivectors,
        use these symbols for the blades.  names should be in the same order
        as bladeTupList.  You may use an empty string for scalars.  By
        default, the name for each non-scalar blade is 'e' plus the indices
        of the blade as given in bladeTupList.

        Example:
          names = ['', 's0', 's1', 'i']  # 2-D


    Attributes
    ----------

    dims :
        dimensionality of vectors (== len(signature))
    sig :
        normalized signature (i.e. all values are +1 or -1)
    firstIdx :
        starting point for vector indices
    bladeTupList :
        list of blades
    gradeList :
        corresponding list of the grades of each blade
    gaDims :
        2**dims
    einf :
        if conformal returns einf
    eo :
        if conformal returns eo
    names :
        pretty-printing symbols for the blades
    even :
        dictionary of even permutations of blades to the canonical blades
    odd :
        dictionary of odd permutations of blades to the canonical blades
    gmt :
        multiplication table for geometric product [1]
    imt :
        multiplication table for inner product [1]
    omt :
        multiplication table for outer product [1]
    lcmt :
        multiplication table for the left-contraction [1]

    [1] The multiplication tables are NumPy arrays of rank 3 with indices like
        the tensor g_ijk discussed above.
    """

    def __init__(self, sig, bladeTupList, firstIdx=0, names=None):
        self.dims = len(sig)
        self.sig = np.array(sig).astype(int)
        self.firstIdx = firstIdx

        self.bladeTupList = list(map(tuple, bladeTupList))
        self._checkList()

        self.gaDims = len(self.bladeTupList)
        self.gradeList = list(map(len, self.bladeTupList))

        self._metric = None

        self.isconformal = False
        self.einf = None
        self.eo = None

        if names is None or isinstance(names, str):
            if isinstance(names, str):
                e = str(names)
            else:
                e = 'e'
            self.names = []

            for i in range(self.gaDims):
                if self.gradeList[i] >= 1:
                    self.names.append(e + ''.join(
                        map(str, self.bladeTupList[i])))
                else:
                    self.names.append('')

        elif len(names) == self.gaDims:
            self.names = names
        else:
            raise ValueError(
                "names list of length %i needs to be of length %i" %
                (len(names), self.gaDims))

        self._genTables()
        self.adjoint_func = get_adjoint_function(self.gradeList)
        self.right_complement_func = self.gen_right_complement_func()
        self.dual_func = self.gen_dual_func()
        self.vee_func = self.gen_vee_func()

    def __hash__(self):
        """ hashs the signature of the layout """
        return hash(tuple(self.sig))

    def gen_dual_func(self):
        """ Generates the dual function for the pseudoscalar """
        if 0 in self.sig:
            # We are degenerate, use the right complement
            return self.right_complement_func
        else:
            Iinv = self.pseudoScalar.inv().value
            gmt_func = self.gmt_func
            @numba.njit
            def dual_func(Xval):
                return gmt_func(Xval, Iinv)
            return dual_func

    def gen_vee_func(self):
        """
        Generates the vee product function
        """
        dual_func = self.dual_func
        omt_func = self.omt_func
        @numba.njit
        def vee(aval, bval):
            return dual_func(omt_func(dual_func(aval), dual_func(bval)))
        return vee

    @property
    def basis_names(self):
        return np.array(list(sorted(self.basis_vectors.keys())), dtype=bytes)

    def dict_to_multivector(self, dict_in):
        """ Takes a dictionary of coefficient values and converts it into a MultiVector object """
        constructed_values = np.zeros(self.gaDims)
        for k in list(dict_in.keys()):
            constructed_values[int(k)] = dict_in[k]
        return self._newMV(constructed_values)

    def __repr__(self):
        s = ("Layout(%r, %r, firstIdx=%r, names=%r)" % (
                list(self.sig),
                self.bladeTupList, self.firstIdx, self.names))
        return s

    def __eq__(self, other):
        if other is self:
            return True
        elif isinstance(other, Layout):
            return np.array_equal(self.sig, other.sig)
        else:
            return NotImplemented

    def __ne__(self, other):
        if other is self:
            return False
        elif isinstance(other, Layout):
            return not np.array_equal(self.sig, other.sig)
        else:
            return NotImplemented

    def parse_multivector(self, mv_string: str) -> 'cf.MultiVector':
        """ Parses a multivector string into a MultiVector object """
        # Get the names of the canonical blades
        blade_name_index_map = {name: index for index, name in enumerate(self.names)}

        # Clean up the input string a bit
        cleaned_string = re.sub('[()]', '', mv_string)

        # Create a multivector
        mv_out = cf.MultiVector(self)

        # Apply the regex
        for m in _blade_pattern.finditer(cleaned_string):
            # Clean up the search result
            cleaned_match = m.group(0)

            # Split on the '^'
            stuff = cleaned_match.split('^')

            if len(stuff) == 2:
                # Extract the value of the blade and the index of the blade
                blade_val = float("".join(stuff[0].split()))
                blade_index = blade_name_index_map[stuff[1].strip()]
                mv_out[blade_index] = blade_val
            elif len(stuff) == 1:
                # Extract the value of the scalar
                blade_val = float("".join(stuff[0].split()))
                blade_index = 0
                mv_out[blade_index] = blade_val
        return mv_out

    def _checkList(self):
        "Ensure validity of arguments."

        # check for uniqueness
        for blade in self.bladeTupList:
            if self.bladeTupList.count(blade) != 1:
                raise ValueError("blades not unique")

        # check for right dimensionality
        if len(self.bladeTupList) != 2**self.dims:
            raise ValueError("incorrect number of blades")

        # check for valid ranges of indices
        valid = list(range(self.firstIdx, self.firstIdx + self.dims))
        try:
            for blade in self.bladeTupList:
                for idx in blade:
                    if (idx not in valid) or (list(blade).count(idx) != 1):
                        raise ValueError()
        except (ValueError, TypeError):
            raise ValueError("invalid bladeTupList; must be a list of tuples")

    def _genTables(self):
        "Generate the multiplication tables."

        self.bladeTupMap = generate_blade_tup_map(self.bladeTupList)
        self.bitmap_to_linear_map = generate_bitmap_to_linear_index_map(self.bladeTupList, self.firstIdx)
        self.linear_map_to_bitmap = np.zeros(len(self.bladeTupMap), dtype=int)
        for bitmap, linear in enumerate(self.bitmap_to_linear_map):
            self.linear_map_to_bitmap[linear] = int(bitmap)

        self.gmt, imt_prod_mask, omt_prod_mask, lcmt_prod_mask = construct_tables(
            np.array(self.gradeList),
            self.linear_map_to_bitmap,
            self.bitmap_to_linear_map,
            self.sig
        )
        self.omt = sparse.where(omt_prod_mask, self.gmt, self.gmt.dtype.type(0))
        self.imt = sparse.where(imt_prod_mask, self.gmt, self.gmt.dtype.type(0))
        self.lcmt = sparse.where(lcmt_prod_mask, self.gmt, self.gmt.dtype.type(0))

        # This generates the functions that will perform the various products
        self.gmt_func = get_mult_function(self.gmt, self.gradeList)
        self.imt_func = get_mult_function(self.imt, self.gradeList)
        self.omt_func = get_mult_function(self.omt, self.gradeList)
        self.lcmt_func = get_mult_function(self.lcmt, self.gradeList)
        self.inv_func = get_leftLaInv(self.gmt, self.gradeList)

        # these are probably not useful, but someone might want them
        self.imt_prod_mask = imt_prod_mask
        self.omt_prod_mask = omt_prod_mask
        self.lcmt_prod_mask = lcmt_prod_mask

    def gmt_func_generator(self, grades_a=None, grades_b=None, filter_mask=None):
        return get_mult_function(
            self.gmt, self.gradeList,
            grades_a=grades_a, grades_b=grades_b, filter_mask=filter_mask
        )

    def imt_func_generator(self, grades_a=None, grades_b=None, filter_mask=None):
        return get_mult_function(
            self.gmt, self.gradeList,
            grades_a=grades_a, grades_b=grades_b, filter_mask=filter_mask
        )

    def omt_func_generator(self, grades_a=None, grades_b=None, filter_mask=None):
        return get_mult_function(
            self.gmt, self.gradeList,
            grades_a=grades_a, grades_b=grades_b, filter_mask=filter_mask
        )

    def lcmt_func_generator(self, grades_a=None, grades_b=None, filter_mask=None):
        return get_mult_function(
            self.gmt, self.gradeList,
            grades_a=grades_a, grades_b=grades_b, filter_mask=filter_mask
        )

    def get_grade_projection_matrix(self, grade):
        """
        Returns the matrix M_g that performs grade projection via left multiplication
        eg. M_g@A.value = A(g).value
        """
        diag_mask = 1.0 * (np.array(self.gradeList) == grade)
        return np.diag(diag_mask)

    def gen_right_complement_func(self):
        """
        Generates the right complement of a multivector
        """
        dims = self.gaDims
        bl = self.blades_list
        signlist = np.zeros(self.gaDims)
        for n in range(len(bl)):
            i = bl[n]
            j = bl[dims-1-n]
            signval = (-1)**((i^j).value[-1] < 0.001)
            signlist[n] = signval

        @numba.njit
        def right_comp_func(Xval):
            Yval = np.zeros(dims)
            for i, s in enumerate(signlist):
                Yval[i] = Xval[dims-1-i]*s
            return Yval
        return right_comp_func

    def get_left_gmt_matrix(self, x):
        """
        This produces the matrix X that performs left multiplication with x
        eg. X@b == (x*b).value
        """
        return val_get_left_gmt_matrix(self.gmt, x.value)

    def get_right_gmt_matrix(self, x):
        """
        This produces the matrix X that performs right multiplication with x
        eg. X@b == (b*x).value
        """
        return val_get_right_gmt_matrix(self.gmt, x.value)

    def MultiVector(self, *args, **kw):
        '''
        create a multivector in this layout

        convenience func to Multivector(layout)
        '''
        return cf.MultiVector(layout=self, *args, **kw)

    def load_ga_file(self, filename):
        """
        Takes a ga file
        Checks it is the same signature as this layout
        Loads the data into an MVArray
        """
        data_array, metric, basis_names, support = read_ga_file(filename)
        if not np.allclose(np.diagonal(metric), self.sig):
            raise ValueError('The signature of the ga file does not match this layout')
        return cf.MVArray.from_value_array(self, data_array)

    def grade_mask(self, grade):
        return np.equal(grade, self.gradeList)

    @property
    def rotor_mask(self):
        return sum(
            self.grade_mask(i)
            for i in range(self.dims + 1)
            if not i % 2
        )

    @property
    def metric(self):
        if self._metric is None:
            self._metric = np.zeros((len(self.basis_vectors), len(self.basis_vectors)))
            for i, v in enumerate(self.basis_vectors_lst):
                for j, v2 in enumerate(self.basis_vectors_lst):
                    self._metric[i, j] = (v | v2)[0]
            return self._metric.copy()
        else:
            return self._metric.copy()

    @property
    def scalar(self):
        '''
        the scalar of value 1, for this GA (a MultiVector object)

        useful for forcing a MultiVector type
        '''
        return self.MultiVector() + 1

    @property
    def pseudoScalar(self):
        '''
        the psuedoScalar
        '''
        return self.blades_list[-1]

    I = pseudoScalar

    def randomMV(self, n=1, **kw):
        '''
        Convenience method to create a random multivector.

        see `clifford.randomMV` for details
        '''
        kw.update(dict(n=n))
        return cf.randomMV(layout=self, **kw)

    def randomV(self, n=1, **kw):
        '''
        generate n random 1-vector s

        '''
        kw.update(dict(n=n, grades=[1]))
        return cf.randomMV(layout=self, **kw)

    def randomRotor(self):
        '''
        generate a random Rotor.

        this is created by muliplying an N unit vectors, where N is
        the dimension of the algebra if its even; else its one less.

        '''
        n = self.dims if self.dims % 2 == 0 else self.dims - 1
        R = reduce(cf.gp, self.randomV(n, normed=True))
        return R

    @property
    def basis_vectors(self):
        return cf.basis_vectors(self)

    @property
    def basis_vectors_lst(self):
        d = self.basis_vectors
        return [d[k] for k in sorted(d.keys())]

    def blades_of_grade(self, grade: int) -> List['MultiVector']:
        '''
        return all blades of a given grade,

        Parameters
        ------------
        grade: int
            the desired grade

        Returns
        --------
        blades : list of MultiVectors
        '''
        return [k for k in self.blades_list if k.grades() == {grade}]

    @property
    def blades_list(self) -> List['MultiVector']:
        '''
        List of blades in this layout matching the order of `self.bladeTupList`
        '''
        blades = self.blades
        return [blades[n] for n in self.names]

    @property
    def blades(self):
        return self.bases()

    def bases(self, *args, **kw):
        '''
        Returns a dictionary mapping basis element names to their MultiVector
        instances, optionally for specific grades

        if you are lazy,  you might do this to populate your namespace
        with the variables of a given layout.

        >>> locals().update(layout.bases())



        See Also
        ---------
        bases
        '''
        return cf.bases(layout=self, *args, **kw)




from clifford import Cl
import unittest
import math
import numpy as np


class PGATests(unittest.TestCase):

    def setUp(self):
        self.algebras = [Cl(3, 0, 1, firstIdx=0), Cl(4, 0, 1, firstIdx=0)]

    def test_integration(self):
        layout = self.algebras[0][0]
        blades = self.algebras[0][1]
        print(blades)

        # A rotor (Euclidean line) and translator (Ideal line)
        def rotor(angle, line):
            return math.cos(angle/2.0) + math.sin(angle/2.0)*line.normal()

        def translator(dist, line):
            return 1.0 + dist/2.0*line

        # PGA is plane based. Vectors are planes. (think linear functionals)
        E0 = blades['e0']          # ideal plane
        E1 = blades['e1']           # x=0 plane
        E2 = blades['e2']           # y=0 plane
        E3 = blades['e3']           # z=0 plane

        # A plane is defined using its homogenous equation ax + by + cz + d = 0
        def PLANE(a, b, c, d):
            return a*E1 + b*E2 + c*E3 + d*E0

        # PGA points are trivectors.
        E123 = E1^E2^E3
        E032 = E0^E3^E2
        E013 = E0^E1^E3
        E021 = E0^E2^E1

        # A point is just a homogeneous point, euclidean coordinates plus the origin
        def POINT(x, y, z):
            return E123 + x*E032 + y*E013 + z*E021

        # for our toy problem (generate points on the surface of a torus)
        # we start with a function that generates motors.
        # circle(t) with t going from 0 to 1.
        def CIRCLE(t, radius, line):
            return rotor(t*math.pi*2.0, line)*translator(radius, E1*E0)

        # a torus is now the product of two circles.
        def TORUS(s, t, r1, l1, r2, l2):
            return CIRCLE(s, r2, l2)*CIRCLE(t, r1, l1)

        # sample the torus points by sandwich with the origin
        def POINT_ON_TORUS(s, t):
            to = TORUS(s, t, 0.25, E1*E2, 0.6, E1*E3)
            return to * E123 * ~to

        # Elements of the even subalgebra (scalar + bivector + pss) of unit length are motors
        ROT = rotor(math.pi/2.0, E1*E2)

        # The outer product ^ is the MEET. Here we intersect the yz (x=0) and xz (y=0) planes.
        AXZ = E1 ^ E2                # x=0, y=0 -> z-axis line

        # line and plane meet in point. We intersect the line along the z-axis (x=0, y=0) with the xy (z=0) plane.
        ORIG = AXZ ^ E3              # x=0, y=0, z=0 -> origin

        # We can also easily create points and join them into a line using the regressive (vee, &) product.
        PX = POINT(1, 0, 0)
        LINE = (ORIG.dual()^PX.dual()).dual()             # & = regressive product, JOIN, here, x-axis line.

        # Lets also create the plane with equation 2x + z - 3 = 0
        P = PLANE(2, 0, 1, -3)

        # rotations work on all elements ..
        ROTATED_LINE = ROT * LINE * ~ROT
        ROTATED_POINT = ROT * PX * ~ROT
        ROTATED_PLANE = ROT * P * ~ROT

        # See the 3D PGA Cheat sheet for a huge collection of useful formulas
        POINT_ON_PLANE = (P | PX) * P

        # output some numbers.
        print("a point       :", str(PX))
        print("a line        :", str(LINE))
        print("a plane       :", str(P))
        print("a rotor       :", str(ROT))
        print("rotated line  :", str(ROTATED_LINE))
        print("rotated point :", str(ROTATED_POINT))
        print("rotated plane :", str(ROTATED_PLANE))
        print("point on plane:", str(POINT_ON_PLANE.normal()))
        print("point on torus:", str(POINT_ON_TORUS(0.0, 0.0)))


if __name__ == '__main__':
    unittest.main()



import itertools
import time

import numpy as np
import pytest

from clifford import Cl, conformalize


class TestInitialisation:

    def test_speed(self):
        algebras = range(2, 9)
        print()  # So that the first number is on a new line
        for i in algebras:
            t_start = time.time()
            Cl(i)
            t_end = time.time()
            print(i, t_end - t_start)

    @pytest.mark.parametrize(
        'algebra',
        [Cl(i) for i in [4]] + [conformalize(Cl(3)[0])],
        ids=['Cl(4)', 'conformalize(Cl(3))']
    )
    def test_sparse_multiply(self, algebra):
        layout = algebra[0]
        # Make two random multivectors
        a = layout.randomMV()
        b = layout.randomMV()
        # Project the multivectors to the grades required
        grades_possibilities = []
        for r in range(1, len(layout.sig)):
            possible_grades = [list(m) for m in list(itertools.combinations(range(len(layout.sig)), r))]
            grades_possibilities += possible_grades
        for i, grades_a in enumerate(grades_possibilities):
            sparse_mv_a = sum([a(k) for k in grades_a])
            for j, grades_b in enumerate(grades_possibilities):
                sparse_mv_b = sum([b(k) for k in grades_b])
                # Compute results
                gp = layout.gmt_func_generator(grades_a=grades_a, grades_b=grades_b)
                result_sparse = gp(sparse_mv_a.value, sparse_mv_b.value)
                result_dense = (sparse_mv_a*sparse_mv_b).value
                # Check they are the same
                np.testing.assert_almost_equal(result_sparse, result_dense)
                print(j+i*len(grades_possibilities), len(grades_possibilities)**2)




import scipy.spatial
import numpy as np
from .. import MVArray
import itertools
import scipy.special
import random
import functools


class GADelaunay(scipy.spatial.Delaunay):
    """
    Uses scipy to produce an hull_dims dimensional Delaunay triangulation of the input
    points

    if hull_dims is not provided it will use the dimension of the input layout
    if hull_dims is provided it will use the first hull_dims basis vectors
    """
    def __init__(self, points, hull_dims=None, furthest_site=False,
                 incremental=False, qhull_options=None):

        # Prep the GA specific stuff
        self.layout = points[0].layout
        if hull_dims is None:
            hull_dims = self.layout.dims
        self.algebra_dims = self.layout.dims
        self.hull_dims = hull_dims
        point_array = np.array([p.value[1:hull_dims+1] for p in points])

        # Run the superclass constructor
        scipy.spatial.Delaunay.__init__(self, point_array,
                                        furthest_site=furthest_site,
                                        incremental=incremental,
                                        qhull_options=qhull_options)
        # Keep the GAversion
        self.GApoints = points

    def conformal_facets(self):
        """
        Returns the list of conformal points in each facet
        """
        if self.layout.isconformal:
            return [[self.GApoints[i] for i in s] for s in self.simplices]
        else:
            raise ValueError('Input points do not seem to be from a conformal algebra')


class GAConvexHull(scipy.spatial.ConvexHull):
    """
    Uses scipy to produce an hull_dims dimensional convex hull of the input points
    if hull_dims is not provided it will use the dimension of the input layout
    if hull_dims is provided it will use the first hull_dims basis vectors
    """
    def __init__(self, points, hull_dims=None, incremental=False, qhull_options=None):

        # Prep the GA specific stuff
        self.layout = points[0].layout
        if hull_dims is None:
            hull_dims = self.layout.dims
        self.algebra_dims = self.layout.dims
        self.hull_dims = hull_dims
        point_array = np.array([p.value[1:hull_dims+1] for p in points])

        # Run the superclass constructor
        scipy.spatial.ConvexHull.__init__(self, point_array,
                                          incremental=incremental,
                                          qhull_options=qhull_options)
        # Keep the GAversion
        self.GApoints = points

    def conformal_facets(self):
        """
        Returns the list of conformal points in each facet
        """
        if self.layout.isconformal:
            return [[self.GApoints[i] for i in s] for s in self.simplices]
        else:
            raise ValueError('Input points do not seem to be from a conformal algebra')

    def conformal_rounds(self):
        """
        Returns the conformal rounds made of the wedge
        product of the edge simplices
        """
        if self.layout.isconformal:
            return [MVArray([self.GApoints[i] for i in s]).op().normal() for s in self.simplices]
        else:
            raise ValueError('Input points do not seem to be from a conformal algebra')

    def conformal_flats(self):
        """
        Returns the conformal flats made of the wedge
        product of the edge simplices with einf
        """
        if self.layout.isconformal:
            return [(r^self.layout.einf).normal() for r in self.conformal_rounds()]
        else:
            raise ValueError('Input points do not seem to be from a conformal algebra')


def naiive_subspace_detector(point_list, grade,
                             point_cost_threshold=0.1,
                             flat_flag=True, n_tests=1000):
    """
    Given a point cloud extract objects of the given grade
    This only works for conformal algebras at the moment
    """
    layout = point_list[0].layout
    if not layout.isconformal:
        raise ValueError('The naiive_subspace_detector only operates on conformal algebras at the moment')
    einf = layout.einf

    def sample_objects(point_list, grade, n_tests, flat_flag=True):
        if flat_flag:
            n_sample_objects = grade - 1
        else:
            n_sample_objects = grade

        # How many combos are there
        nchoosek = scipy.special.comb(len(point_list), n_sample_objects)

        # Get a combo generator
        combos = itertools.combinations(point_list, n_sample_objects)

        # Sample from the generator
        n_samples = int(min([n_tests, nchoosek]))
        sampled_objects_list = random.sample(list(combos), n_samples)

        # Wedge the results
        output_list = []
        for i in range(n_samples):
            print(i, flush=True)
            sampled_objects = sampled_objects_list[i]
            wedged_points = functools.reduce(lambda x, y: x ^ y, sampled_objects)
            output_object = wedged_points
            if flat_flag:
                output_object = output_object ^ einf
            final_object = (output_object).normal()
            output_list.append(final_object)
        return output_list

    def associate_points_with_objects(point_list, object_list, point_cost_threshold=0.1):
        # Check how close all the points are to each object
        point_associations = []
        for obj in object_list:
            this_obj_association = []
            for i, p in enumerate(point_list):
                if abs(obj ^ p) < point_cost_threshold:
                    this_obj_association.append(i)
            point_associations.append(this_obj_association)
        return point_associations

    print('SAMPLING', flush=True)
    object_list = sample_objects(point_list, grade, n_tests, flat_flag=flat_flag)

    print('ASSOCIATING', flush=True)
    point_associations = associate_points_with_objects(point_list,
                                                       object_list,
                                                       point_cost_threshold=point_cost_threshold)
    return point_associations, object_list



import numpy as np
from numba import njit, cuda


@cuda.jit(device=True)
def gmt_func_o0(value, other_value):
    return 0 + 1.0*value[0]*other_value[0] + 1.0*value[3]*other_value[3] + 1.0*value[4]*other_value[4] + -1.0*value[5]*other_value[5] + -1.0*value[6]*other_value[6] + -1.0*value[7]*other_value[7] + -1.0*value[8]*other_value[8] + 1.0*value[9]*other_value[9] + -1.0*value[10]*other_value[10] + -1.0*value[11]*other_value[11] + 1.0*value[12]*other_value[12] + -1.0*value[13]*other_value[13] + 1.0*value[14]*other_value[14] + 1.0*value[15]*other_value[15] + 1.0*value[2]*other_value[2] + -1.0*value[16]*other_value[16] + 1.0*value[18]*other_value[18] + -1.0*value[19]*other_value[19] + 1.0*value[20]*other_value[20] + 1.0*value[21]*other_value[21] + -1.0*value[22]*other_value[22] + 1.0*value[23]*other_value[23] + 1.0*value[24]*other_value[24] + 1.0*value[25]*other_value[25] + 1.0*value[26]*other_value[26] + -1.0*value[27]*other_value[27] + -1.0*value[28]*other_value[28] + -1.0*value[29]*other_value[29] + -1.0*value[30]*other_value[30] + -1.0*value[17]*other_value[17] + 1.0*value[1]*other_value[1] + -1.0*value[31]*other_value[31]

@cuda.jit(device=True)
def gmt_func_o1(value, other_value):
    return 0 + 1.0*value[1]*other_value[0] + -1.0*value[23]*other_value[27] + 1.0*value[20]*other_value[14] + 1.0*value[14]*other_value[20] + 1.0*value[18]*other_value[12] + -1.0*value[30]*other_value[31] + 1.0*value[0]*other_value[1] + 1.0*value[22]*other_value[26] + -1.0*value[10]*other_value[16] + 1.0*value[29]*other_value[25] + -1.0*value[2]*other_value[6] + -1.0*value[13]*other_value[19] + -1.0*value[3]*other_value[7] + -1.0*value[11]*other_value[17] + -1.0*value[19]*other_value[13] + 1.0*value[21]*other_value[15] + -1.0*value[4]*other_value[8] + 1.0*value[28]*other_value[24] + -1.0*value[9]*other_value[5] + 1.0*value[12]*other_value[18] + 1.0*value[5]*other_value[9] + -1.0*value[17]*other_value[11] + -1.0*value[26]*other_value[22] + -1.0*value[25]*other_value[29] + 1.0*value[7]*other_value[3] + -1.0*value[16]*other_value[10] + 1.0*value[6]*other_value[2] + -1.0*value[31]*other_value[30] + -1.0*value[24]*other_value[28] + 1.0*value[15]*other_value[21] + 1.0*value[8]*other_value[4] + 1.0*value[27]*other_value[23]

@cuda.jit(device=True)
def gmt_func_o2(value, other_value):
    return 0 + 1.0*value[11]*other_value[4] + 1.0*value[29]*other_value[31] + 1.0*value[5]*other_value[12] + -1.0*value[27]*other_value[20] + 1.0*value[20]*other_value[27] + -1.0*value[12]*other_value[5] + 1.0*value[0]*other_value[2] + 1.0*value[24]*other_value[15] + -1.0*value[6]*other_value[1] + 1.0*value[26]*other_value[19] + 1.0*value[10]*other_value[3] + -1.0*value[25]*other_value[30] + -1.0*value[9]*other_value[18] + 1.0*value[23]*other_value[14] + 1.0*value[7]*other_value[16] + -1.0*value[4]*other_value[11] + 1.0*value[8]*other_value[17] + -1.0*value[28]*other_value[21] + 1.0*value[21]*other_value[28] + -1.0*value[22]*other_value[13] + -1.0*value[3]*other_value[10] + 1.0*value[14]*other_value[23] + 1.0*value[30]*other_value[25] + -1.0*value[13]*other_value[22] + -1.0*value[18]*other_value[9] + 1.0*value[15]*other_value[24] + 1.0*value[1]*other_value[6] + 1.0*value[17]*other_value[8] + -1.0*value[19]*other_value[26] + 1.0*value[2]*other_value[0] + 1.0*value[16]*other_value[7] + 1.0*value[31]*other_value[29]

@cuda.jit(device=True)
def gmt_func_o3(value, other_value):
    return 0 + -1.0*value[9]*other_value[20] + -1.0*value[23]*other_value[12] + 1.0*value[0]*other_value[3] + 1.0*value[17]*other_value[26] + -1.0*value[4]*other_value[13] + 1.0*value[8]*other_value[19] + -1.0*value[28]*other_value[31] + -1.0*value[14]*other_value[5] + 1.0*value[24]*other_value[30] + 1.0*value[25]*other_value[15] + 1.0*value[15]*other_value[25] + -1.0*value[31]*other_value[28] + -1.0*value[7]*other_value[1] + -1.0*value[6]*other_value[16] + -1.0*value[26]*other_value[17] + 1.0*value[1]*other_value[7] + -1.0*value[16]*other_value[6] + 1.0*value[27]*other_value[18] + 1.0*value[5]*other_value[14] + -1.0*value[10]*other_value[2] + -1.0*value[18]*other_value[27] + -1.0*value[12]*other_value[23] + 1.0*value[22]*other_value[11] + -1.0*value[30]*other_value[24] + 1.0*value[11]*other_value[22] + 1.0*value[3]*other_value[0] + 1.0*value[2]*other_value[10] + 1.0*value[21]*other_value[29] + -1.0*value[29]*other_value[21] + 1.0*value[19]*other_value[8] + 1.0*value[13]*other_value[4] + -1.0*value[20]*other_value[9]

@cuda.jit(device=True)
def gmt_func_o4(value, other_value):
    return 0 + 1.0*value[0]*other_value[4] + -1.0*value[17]*other_value[6] + 1.0*value[2]*other_value[11] + -1.0*value[18]*other_value[28] + -1.0*value[21]*other_value[9] + -1.0*value[25]*other_value[14] + -1.0*value[7]*other_value[19] + 1.0*value[31]*other_value[27] + 1.0*value[27]*other_value[31] + 1.0*value[5]*other_value[15] + -1.0*value[13]*other_value[3] + -1.0*value[16]*other_value[26] + -1.0*value[6]*other_value[17] + 1.0*value[30]*other_value[23] + -1.0*value[20]*other_value[29] + 1.0*value[1]*other_value[8] + 1.0*value[26]*other_value[16] + -1.0*value[8]*other_value[1] + -1.0*value[15]*other_value[5] + 1.0*value[3]*other_value[13] + 1.0*value[28]*other_value[18] + -1.0*value[9]*other_value[21] + -1.0*value[12]*other_value[24] + -1.0*value[11]*other_value[2] + -1.0*value[23]*other_value[30] + 1.0*value[29]*other_value[20] + -1.0*value[14]*other_value[25] + 1.0*value[4]*other_value[0] + -1.0*value[24]*other_value[12] + -1.0*value[22]*other_value[10] + -1.0*value[19]*other_value[7] + -1.0*value[10]*other_value[22]

@cuda.jit(device=True)
def gmt_func_o5(value, other_value):
    return 0 + -1.0*value[19]*other_value[29] + -1.0*value[22]*other_value[30] + -1.0*value[14]*other_value[3] + -1.0*value[23]*other_value[10] + 1.0*value[0]*other_value[5] + -1.0*value[20]*other_value[7] + -1.0*value[12]*other_value[2] + -1.0*value[6]*other_value[18] + -1.0*value[11]*other_value[24] + -1.0*value[18]*other_value[6] + -1.0*value[16]*other_value[27] + -1.0*value[9]*other_value[1] + -1.0*value[21]*other_value[8] + -1.0*value[7]*other_value[20] + -1.0*value[8]*other_value[21] + -1.0*value[25]*other_value[13] + -1.0*value[17]*other_value[28] + -1.0*value[13]*other_value[25] + -1.0*value[15]*other_value[4] + -1.0*value[10]*other_value[23] + 1.0*value[26]*other_value[31] + -1.0*value[24]*other_value[11] + 1.0*value[27]*other_value[16] + 1.0*value[30]*other_value[22] + 1.0*value[3]*other_value[14] + 1.0*value[1]*other_value[9] + 1.0*value[31]*other_value[26] + 1.0*value[5]*other_value[0] + 1.0*value[4]*other_value[15] + 1.0*value[29]*other_value[19] + 1.0*value[28]*other_value[17] + 1.0*value[2]*other_value[12]

@cuda.jit(device=True)
def gmt_func_o6(value, other_value):
    return 0 + 1.0*value[11]*other_value[8] + 1.0*value[3]*other_value[16] + 1.0*value[10]*other_value[7] + -1.0*value[26]*other_value[13] + -1.0*value[29]*other_value[30] + 1.0*value[17]*other_value[4] + -1.0*value[19]*other_value[22] + 1.0*value[1]*other_value[2] + 1.0*value[25]*other_value[31] + 1.0*value[4]*other_value[17] + -1.0*value[12]*other_value[9] + 1.0*value[0]*other_value[6] + 1.0*value[22]*other_value[19] + 1.0*value[20]*other_value[23] + 1.0*value[21]*other_value[24] + -1.0*value[7]*other_value[10] + -1.0*value[23]*other_value[20] + 1.0*value[6]*other_value[0] + 1.0*value[28]*other_value[15] + 1.0*value[31]*other_value[25] + -1.0*value[5]*other_value[18] + -1.0*value[18]*other_value[5] + -1.0*value[2]*other_value[1] + 1.0*value[9]*other_value[12] + 1.0*value[27]*other_value[14] + 1.0*value[30]*other_value[29] + -1.0*value[24]*other_value[21] + -1.0*value[13]*other_value[26] + 1.0*value[14]*other_value[27] + -1.0*value[8]*other_value[11] + 1.0*value[16]*other_value[3] + 1.0*value[15]*other_value[28]

@cuda.jit(device=True)
def gmt_func_o7(value, other_value):
    return 0 + -1.0*value[25]*other_value[21] + 1.0*value[4]*other_value[19] + -1.0*value[20]*other_value[5] + 1.0*value[21]*other_value[25] + -1.0*value[8]*other_value[13] + 1.0*value[11]*other_value[26] + -1.0*value[24]*other_value[31] + -1.0*value[3]*other_value[1] + -1.0*value[16]*other_value[2] + 1.0*value[28]*other_value[30] + 1.0*value[17]*other_value[22] + -1.0*value[31]*other_value[24] + -1.0*value[18]*other_value[23] + -1.0*value[14]*other_value[9] + 1.0*value[6]*other_value[10] + 1.0*value[15]*other_value[29] + 1.0*value[19]*other_value[4] + -1.0*value[30]*other_value[28] + -1.0*value[12]*other_value[27] + -1.0*value[10]*other_value[6] + 1.0*value[26]*other_value[11] + 1.0*value[23]*other_value[18] + -1.0*value[5]*other_value[20] + 1.0*value[13]*other_value[8] + -1.0*value[2]*other_value[16] + 1.0*value[1]*other_value[3] + 1.0*value[0]*other_value[7] + 1.0*value[7]*other_value[0] + -1.0*value[22]*other_value[17] + 1.0*value[9]*other_value[14] + 1.0*value[29]*other_value[15] + -1.0*value[27]*other_value[12]

@cuda.jit(device=True)
def gmt_func_o8(value, other_value):
    return 0 + -1.0*value[10]*other_value[26] + 1.0*value[9]*other_value[15] + -1.0*value[17]*other_value[2] + -1.0*value[29]*other_value[14] + 1.0*value[24]*other_value[18] + 1.0*value[23]*other_value[31] + 1.0*value[22]*other_value[16] + -1.0*value[3]*other_value[19] + 1.0*value[0]*other_value[8] + -1.0*value[20]*other_value[25] + -1.0*value[4]*other_value[1] + 1.0*value[31]*other_value[23] + -1.0*value[14]*other_value[29] + -1.0*value[18]*other_value[24] + 1.0*value[6]*other_value[11] + -1.0*value[19]*other_value[3] + 1.0*value[30]*other_value[27] + -1.0*value[15]*other_value[9] + -1.0*value[26]*other_value[10] + -1.0*value[13]*other_value[7] + -1.0*value[2]*other_value[17] + -1.0*value[27]*other_value[30] + 1.0*value[1]*other_value[4] + -1.0*value[16]*other_value[22] + -1.0*value[28]*other_value[12] + -1.0*value[5]*other_value[21] + 1.0*value[7]*other_value[13] + 1.0*value[25]*other_value[20] + 1.0*value[8]*other_value[0] + -1.0*value[11]*other_value[6] + -1.0*value[12]*other_value[28] + -1.0*value[21]*other_value[5]

@cuda.jit(device=True)
def gmt_func_o9(value, other_value):
    return 0 + 1.0*value[31]*other_value[22] + -1.0*value[18]*other_value[2] + -1.0*value[14]*other_value[7] + -1.0*value[15]*other_value[8] + -1.0*value[13]*other_value[29] + 1.0*value[22]*other_value[31] + -1.0*value[3]*other_value[20] + 1.0*value[30]*other_value[26] + -1.0*value[21]*other_value[4] + -1.0*value[11]*other_value[28] + 1.0*value[0]*other_value[9] + -1.0*value[10]*other_value[27] + -1.0*value[12]*other_value[6] + -1.0*value[20]*other_value[3] + -1.0*value[19]*other_value[25] + -1.0*value[2]*other_value[18] + -1.0*value[29]*other_value[13] + -1.0*value[27]*other_value[10] + -1.0*value[17]*other_value[24] + 1.0*value[23]*other_value[16] + -1.0*value[4]*other_value[21] + 1.0*value[1]*other_value[5] + 1.0*value[9]*other_value[0] + -1.0*value[26]*other_value[30] + 1.0*value[8]*other_value[15] + -1.0*value[28]*other_value[11] + 1.0*value[25]*other_value[19] + 1.0*value[7]*other_value[14] + 1.0*value[6]*other_value[12] + 1.0*value[24]*other_value[17] + -1.0*value[16]*other_value[23] + -1.0*value[5]*other_value[1]

@cuda.jit(device=True)
def gmt_func_o10(value, other_value):
    return 0 + 1.0*value[22]*other_value[4] + -1.0*value[17]*other_value[19] + -1.0*value[14]*other_value[12] + 1.0*value[21]*other_value[31] + 1.0*value[16]*other_value[1] + 1.0*value[24]*other_value[25] + 1.0*value[13]*other_value[11] + 1.0*value[4]*other_value[22] + 1.0*value[0]*other_value[10] + 1.0*value[12]*other_value[14] + 1.0*value[18]*other_value[20] + 1.0*value[1]*other_value[16] + 1.0*value[7]*other_value[6] + 1.0*value[29]*other_value[28] + -1.0*value[25]*other_value[24] + -1.0*value[6]*other_value[7] + 1.0*value[2]*other_value[3] + 1.0*value[19]*other_value[17] + -1.0*value[26]*other_value[8] + 1.0*value[30]*other_value[15] + -1.0*value[20]*other_value[18] + -1.0*value[3]*other_value[2] + -1.0*value[11]*other_value[13] + 1.0*value[27]*other_value[9] + -1.0*value[28]*other_value[29] + 1.0*value[31]*other_value[21] + 1.0*value[10]*other_value[0] + -1.0*value[8]*other_value[26] + -1.0*value[5]*other_value[23] + 1.0*value[9]*other_value[27] + 1.0*value[15]*other_value[30] + -1.0*value[23]*other_value[5]

@cuda.jit(device=True)
def gmt_func_o11(value, other_value):
    return 0 + 1.0*value[16]*other_value[19] + 1.0*value[7]*other_value[26] + 1.0*value[27]*other_value[29] + -1.0*value[31]*other_value[20] + 1.0*value[26]*other_value[7] + 1.0*value[1]*other_value[17] + -1.0*value[15]*other_value[12] + -1.0*value[20]*other_value[31] + -1.0*value[6]*other_value[8] + 1.0*value[12]*other_value[15] + 1.0*value[9]*other_value[28] + 1.0*value[28]*other_value[9] + 1.0*value[2]*other_value[4] + -1.0*value[30]*other_value[14] + -1.0*value[21]*other_value[18] + 1.0*value[25]*other_value[23] + -1.0*value[23]*other_value[25] + -1.0*value[24]*other_value[5] + -1.0*value[22]*other_value[3] + -1.0*value[13]*other_value[10] + -1.0*value[3]*other_value[22] + -1.0*value[14]*other_value[30] + -1.0*value[5]*other_value[24] + 1.0*value[11]*other_value[0] + 1.0*value[8]*other_value[6] + -1.0*value[4]*other_value[2] + 1.0*value[17]*other_value[1] + 1.0*value[10]*other_value[13] + -1.0*value[19]*other_value[16] + 1.0*value[18]*other_value[21] + -1.0*value[29]*other_value[27] + 1.0*value[0]*other_value[11]

@cuda.jit(device=True)
def gmt_func_o12(value, other_value):
    return 0 + 1.0*value[0]*other_value[12] + 1.0*value[2]*other_value[5] + -1.0*value[23]*other_value[3] + -1.0*value[20]*other_value[16] + -1.0*value[19]*other_value[31] + 1.0*value[16]*other_value[20] + 1.0*value[18]*other_value[1] + 1.0*value[25]*other_value[22] + -1.0*value[5]*other_value[2] + 1.0*value[1]*other_value[18] + -1.0*value[6]*other_value[9] + -1.0*value[24]*other_value[4] + 1.0*value[8]*other_value[28] + -1.0*value[14]*other_value[10] + 1.0*value[10]*other_value[14] + -1.0*value[29]*other_value[26] + 1.0*value[7]*other_value[27] + -1.0*value[31]*other_value[19] + -1.0*value[15]*other_value[11] + 1.0*value[11]*other_value[15] + 1.0*value[9]*other_value[6] + -1.0*value[21]*other_value[17] + 1.0*value[27]*other_value[7] + -1.0*value[22]*other_value[25] + -1.0*value[3]*other_value[23] + -1.0*value[4]*other_value[24] + 1.0*value[28]*other_value[8] + -1.0*value[13]*other_value[30] + 1.0*value[17]*other_value[21] + -1.0*value[30]*other_value[13] + 1.0*value[26]*other_value[29] + 1.0*value[12]*other_value[0]

@cuda.jit(device=True)
def gmt_func_o13(value, other_value):
    return 0 + 1.0*value[0]*other_value[13] + 1.0*value[19]*other_value[1] + 1.0*value[1]*other_value[19] + 1.0*value[31]*other_value[18] + -1.0*value[5]*other_value[25] + -1.0*value[27]*other_value[28] + 1.0*value[18]*other_value[31] + 1.0*value[14]*other_value[15] + -1.0*value[10]*other_value[11] + 1.0*value[20]*other_value[21] + 1.0*value[8]*other_value[7] + 1.0*value[11]*other_value[10] + -1.0*value[21]*other_value[20] + -1.0*value[26]*other_value[6] + -1.0*value[16]*other_value[17] + 1.0*value[3]*other_value[4] + -1.0*value[25]*other_value[5] + 1.0*value[28]*other_value[27] + 1.0*value[30]*other_value[12] + 1.0*value[22]*other_value[2] + 1.0*value[17]*other_value[16] + -1.0*value[15]*other_value[14] + -1.0*value[24]*other_value[23] + -1.0*value[4]*other_value[3] + -1.0*value[6]*other_value[26] + 1.0*value[2]*other_value[22] + 1.0*value[9]*other_value[29] + 1.0*value[13]*other_value[0] + -1.0*value[7]*other_value[8] + 1.0*value[12]*other_value[30] + 1.0*value[23]*other_value[24] + 1.0*value[29]*other_value[9]

@cuda.jit(device=True)
def gmt_func_o14(value, other_value):
    return 0 + -1.0*value[24]*other_value[22] + 1.0*value[29]*other_value[8] + 1.0*value[28]*other_value[26] + -1.0*value[10]*other_value[12] + 1.0*value[17]*other_value[31] + 1.0*value[1]*other_value[20] + 1.0*value[31]*other_value[17] + 1.0*value[8]*other_value[29] + 1.0*value[23]*other_value[2] + 1.0*value[9]*other_value[7] + 1.0*value[22]*other_value[24] + 1.0*value[18]*other_value[16] + 1.0*value[19]*other_value[21] + -1.0*value[21]*other_value[19] + -1.0*value[16]*other_value[18] + -1.0*value[6]*other_value[27] + 1.0*value[13]*other_value[15] + -1.0*value[7]*other_value[9] + 1.0*value[20]*other_value[1] + -1.0*value[5]*other_value[3] + 1.0*value[2]*other_value[23] + -1.0*value[27]*other_value[6] + -1.0*value[15]*other_value[13] + 1.0*value[30]*other_value[11] + 1.0*value[11]*other_value[30] + -1.0*value[4]*other_value[25] + 1.0*value[14]*other_value[0] + -1.0*value[25]*other_value[4] + -1.0*value[26]*other_value[28] + 1.0*value[3]*other_value[5] + 1.0*value[0]*other_value[14] + 1.0*value[12]*other_value[10]

@cuda.jit(device=True)
def gmt_func_o15(value, other_value):
    return 0 + -1.0*value[17]*other_value[18] + -1.0*value[16]*other_value[31] + 1.0*value[0]*other_value[15] + 1.0*value[15]*other_value[0] + 1.0*value[3]*other_value[25] + -1.0*value[29]*other_value[7] + 1.0*value[23]*other_value[22] + 1.0*value[9]*other_value[8] + 1.0*value[4]*other_value[5] + 1.0*value[24]*other_value[2] + 1.0*value[1]*other_value[21] + -1.0*value[8]*other_value[9] + 1.0*value[25]*other_value[3] + -1.0*value[28]*other_value[6] + -1.0*value[6]*other_value[28] + -1.0*value[5]*other_value[4] + -1.0*value[27]*other_value[26] + 1.0*value[26]*other_value[27] + -1.0*value[7]*other_value[29] + -1.0*value[10]*other_value[30] + -1.0*value[22]*other_value[23] + -1.0*value[19]*other_value[20] + 1.0*value[21]*other_value[1] + -1.0*value[30]*other_value[10] + -1.0*value[31]*other_value[16] + 1.0*value[12]*other_value[11] + 1.0*value[20]*other_value[19] + 1.0*value[2]*other_value[24] + 1.0*value[18]*other_value[17] + 1.0*value[14]*other_value[13] + -1.0*value[11]*other_value[12] + -1.0*value[13]*other_value[14]

@cuda.jit(device=True)
def gmt_func_o16(value, other_value):
    return 0 + -1.0*value[27]*other_value[5] + 1.0*value[18]*other_value[14] + 1.0*value[26]*other_value[4] + 1.0*value[31]*other_value[15] + 1.0*value[25]*other_value[28] + 1.0*value[21]*other_value[30] + -1.0*value[24]*other_value[29] + 1.0*value[28]*other_value[25] + 1.0*value[30]*other_value[21] + -1.0*value[20]*other_value[12] + 1.0*value[23]*other_value[9] + -1.0*value[17]*other_value[13] + -1.0*value[22]*other_value[8] + 1.0*value[19]*other_value[11] + -1.0*value[29]*other_value[24] + 1.0*value[15]*other_value[31] + 1.0*value[0]*other_value[16] + 1.0*value[8]*other_value[22] + 1.0*value[12]*other_value[20] + -1.0*value[11]*other_value[19] + 1.0*value[3]*other_value[6] + -1.0*value[7]*other_value[2] + 1.0*value[13]*other_value[17] + -1.0*value[9]*other_value[23] + 1.0*value[16]*other_value[0] + 1.0*value[6]*other_value[3] + -1.0*value[4]*other_value[26] + 1.0*value[5]*other_value[27] + 1.0*value[10]*other_value[1] + -1.0*value[2]*other_value[7] + 1.0*value[1]*other_value[10] + -1.0*value[14]*other_value[18]

@cuda.jit(device=True)
def gmt_func_o17(value, other_value):
    return 0 + 1.0*value[0]*other_value[17] + 1.0*value[3]*other_value[26] + -1.0*value[7]*other_value[22] + -1.0*value[19]*other_value[10] + 1.0*value[12]*other_value[21] + 1.0*value[4]*other_value[6] + 1.0*value[1]*other_value[11] + 1.0*value[5]*other_value[28] + -1.0*value[25]*other_value[27] + 1.0*value[18]*other_value[15] + -1.0*value[26]*other_value[3] + 1.0*value[6]*other_value[4] + -1.0*value[28]*other_value[5] + -1.0*value[27]*other_value[25] + -1.0*value[14]*other_value[31] + -1.0*value[13]*other_value[16] + 1.0*value[29]*other_value[23] + -1.0*value[20]*other_value[30] + -1.0*value[9]*other_value[24] + 1.0*value[23]*other_value[29] + 1.0*value[22]*other_value[7] + -1.0*value[21]*other_value[12] + -1.0*value[30]*other_value[20] + 1.0*value[17]*other_value[0] + 1.0*value[24]*other_value[9] + -1.0*value[2]*other_value[8] + -1.0*value[15]*other_value[18] + 1.0*value[10]*other_value[19] + 1.0*value[16]*other_value[13] + -1.0*value[8]*other_value[2] + -1.0*value[31]*other_value[14] + 1.0*value[11]*other_value[1]

@cuda.jit(device=True)
def gmt_func_o18(value, other_value):
    return 0 + 1.0*value[22]*other_value[29] + 1.0*value[10]*other_value[20] + -1.0*value[13]*other_value[31] + -1.0*value[27]*other_value[3] + 1.0*value[23]*other_value[7] + -1.0*value[26]*other_value[25] + 1.0*value[0]*other_value[18] + 1.0*value[6]*other_value[5] + -1.0*value[19]*other_value[30] + -1.0*value[8]*other_value[24] + 1.0*value[24]*other_value[8] + -1.0*value[21]*other_value[11] + -1.0*value[25]*other_value[26] + -1.0*value[20]*other_value[10] + 1.0*value[11]*other_value[21] + -1.0*value[7]*other_value[23] + -1.0*value[31]*other_value[13] + 1.0*value[12]*other_value[1] + -1.0*value[9]*other_value[2] + 1.0*value[5]*other_value[6] + 1.0*value[16]*other_value[14] + -1.0*value[28]*other_value[4] + 1.0*value[29]*other_value[22] + 1.0*value[4]*other_value[28] + 1.0*value[3]*other_value[27] + 1.0*value[17]*other_value[15] + 1.0*value[1]*other_value[12] + -1.0*value[15]*other_value[17] + 1.0*value[18]*other_value[0] + -1.0*value[30]*other_value[19] + -1.0*value[14]*other_value[16] + -1.0*value[2]*other_value[9]

@cuda.jit(device=True)
def gmt_func_o19(value, other_value):
    return 0 + 1.0*value[7]*other_value[4] + -1.0*value[29]*other_value[5] + 1.0*value[20]*other_value[15] + 1.0*value[25]*other_value[9] + 1.0*value[31]*other_value[12] + 1.0*value[12]*other_value[31] + 1.0*value[26]*other_value[2] + 1.0*value[4]*other_value[7] + 1.0*value[24]*other_value[27] + 1.0*value[13]*other_value[1] + 1.0*value[17]*other_value[10] + -1.0*value[15]*other_value[20] + -1.0*value[21]*other_value[14] + -1.0*value[3]*other_value[8] + 1.0*value[11]*other_value[16] + -1.0*value[23]*other_value[28] + -1.0*value[2]*other_value[26] + 1.0*value[30]*other_value[18] + -1.0*value[9]*other_value[25] + -1.0*value[22]*other_value[6] + -1.0*value[16]*other_value[11] + -1.0*value[8]*other_value[3] + 1.0*value[6]*other_value[22] + 1.0*value[0]*other_value[19] + 1.0*value[19]*other_value[0] + 1.0*value[14]*other_value[21] + 1.0*value[27]*other_value[24] + -1.0*value[10]*other_value[17] + -1.0*value[28]*other_value[23] + 1.0*value[18]*other_value[30] + 1.0*value[1]*other_value[13] + 1.0*value[5]*other_value[29]

@cuda.jit(device=True)
def gmt_func_o20(value, other_value):
    return 0 + -1.0*value[21]*other_value[13] + -1.0*value[3]*other_value[9] + -1.0*value[9]*other_value[3] + 1.0*value[1]*other_value[14] + 1.0*value[25]*other_value[8] + 1.0*value[0]*other_value[20] + 1.0*value[4]*other_value[29] + 1.0*value[24]*other_value[26] + 1.0*value[11]*other_value[31] + -1.0*value[2]*other_value[27] + 1.0*value[13]*other_value[21] + 1.0*value[12]*other_value[16] + 1.0*value[30]*other_value[17] + 1.0*value[7]*other_value[5] + -1.0*value[8]*other_value[25] + 1.0*value[6]*other_value[23] + 1.0*value[26]*other_value[24] + -1.0*value[16]*other_value[12] + -1.0*value[10]*other_value[18] + -1.0*value[15]*other_value[19] + 1.0*value[31]*other_value[11] + 1.0*value[20]*other_value[0] + 1.0*value[5]*other_value[7] + -1.0*value[28]*other_value[22] + 1.0*value[17]*other_value[30] + -1.0*value[22]*other_value[28] + 1.0*value[18]*other_value[10] + -1.0*value[29]*other_value[4] + 1.0*value[19]*other_value[15] + 1.0*value[14]*other_value[1] + -1.0*value[23]*other_value[6] + 1.0*value[27]*other_value[2]

@cuda.jit(device=True)
def gmt_func_o21(value, other_value):
    return 0 + -1.0*value[24]*other_value[6] + -1.0*value[11]*other_value[18] + -1.0*value[26]*other_value[23] + 1.0*value[18]*other_value[11] + -1.0*value[23]*other_value[26] + -1.0*value[9]*other_value[4] + -1.0*value[10]*other_value[31] + -1.0*value[30]*other_value[16] + 1.0*value[27]*other_value[22] + 1.0*value[5]*other_value[8] + 1.0*value[22]*other_value[27] + -1.0*value[2]*other_value[28] + -1.0*value[16]*other_value[30] + 1.0*value[8]*other_value[5] + 1.0*value[0]*other_value[21] + 1.0*value[20]*other_value[13] + 1.0*value[12]*other_value[17] + -1.0*value[19]*other_value[14] + 1.0*value[15]*other_value[1] + -1.0*value[3]*other_value[29] + 1.0*value[28]*other_value[2] + -1.0*value[25]*other_value[7] + 1.0*value[29]*other_value[3] + 1.0*value[7]*other_value[25] + 1.0*value[1]*other_value[15] + 1.0*value[21]*other_value[0] + -1.0*value[4]*other_value[9] + -1.0*value[31]*other_value[10] + 1.0*value[6]*other_value[24] + -1.0*value[17]*other_value[12] + 1.0*value[14]*other_value[19] + -1.0*value[13]*other_value[20]

@cuda.jit(device=True)
def gmt_func_o22(value, other_value):
    return 0 + 1.0*value[28]*other_value[20] + 1.0*value[4]*other_value[10] + 1.0*value[2]*other_value[13] + 1.0*value[10]*other_value[4] + -1.0*value[26]*other_value[1] + -1.0*value[3]*other_value[11] + -1.0*value[27]*other_value[21] + -1.0*value[31]*other_value[9] + 1.0*value[7]*other_value[17] + 1.0*value[25]*other_value[12] + -1.0*value[29]*other_value[18] + 1.0*value[5]*other_value[30] + -1.0*value[6]*other_value[19] + -1.0*value[8]*other_value[16] + -1.0*value[24]*other_value[14] + -1.0*value[30]*other_value[5] + -1.0*value[9]*other_value[31] + 1.0*value[23]*other_value[15] + 1.0*value[0]*other_value[22] + 1.0*value[16]*other_value[8] + -1.0*value[17]*other_value[7] + 1.0*value[20]*other_value[28] + -1.0*value[11]*other_value[3] + -1.0*value[12]*other_value[25] + 1.0*value[13]*other_value[2] + -1.0*value[18]*other_value[29] + 1.0*value[1]*other_value[26] + -1.0*value[15]*other_value[23] + 1.0*value[14]*other_value[24] + 1.0*value[19]*other_value[6] + -1.0*value[21]*other_value[27] + 1.0*value[22]*other_value[0]

@cuda.jit(device=True)
def gmt_func_o23(value, other_value):
    return 0 + 1.0*value[2]*other_value[14] + 1.0*value[4]*other_value[30] + -1.0*value[24]*other_value[13] + -1.0*value[8]*other_value[31] + 1.0*value[0]*other_value[23] + -1.0*value[11]*other_value[25] + 1.0*value[14]*other_value[2] + -1.0*value[21]*other_value[26] + -1.0*value[29]*other_value[17] + -1.0*value[18]*other_value[7] + 1.0*value[25]*other_value[11] + 1.0*value[7]*other_value[18] + 1.0*value[20]*other_value[6] + 1.0*value[19]*other_value[28] + -1.0*value[17]*other_value[29] + -1.0*value[6]*other_value[20] + -1.0*value[26]*other_value[21] + -1.0*value[31]*other_value[8] + -1.0*value[27]*other_value[1] + 1.0*value[13]*other_value[24] + 1.0*value[5]*other_value[10] + -1.0*value[12]*other_value[3] + 1.0*value[28]*other_value[19] + 1.0*value[22]*other_value[15] + 1.0*value[10]*other_value[5] + -1.0*value[30]*other_value[4] + 1.0*value[16]*other_value[9] + -1.0*value[3]*other_value[12] + 1.0*value[1]*other_value[27] + -1.0*value[15]*other_value[22] + -1.0*value[9]*other_value[16] + 1.0*value[23]*other_value[0]

@cuda.jit(device=True)
def gmt_func_o24(value, other_value):
    return 0 + -1.0*value[3]*other_value[30] + -1.0*value[4]*other_value[12] + 1.0*value[1]*other_value[28] + 1.0*value[7]*other_value[31] + -1.0*value[19]*other_value[27] + 1.0*value[0]*other_value[24] + 1.0*value[5]*other_value[11] + 1.0*value[10]*other_value[25] + 1.0*value[20]*other_value[26] + 1.0*value[14]*other_value[22] + 1.0*value[26]*other_value[20] + -1.0*value[12]*other_value[4] + 1.0*value[15]*other_value[2] + -1.0*value[25]*other_value[10] + -1.0*value[28]*other_value[1] + 1.0*value[29]*other_value[16] + 1.0*value[8]*other_value[18] + -1.0*value[6]*other_value[21] + 1.0*value[16]*other_value[29] + 1.0*value[30]*other_value[3] + 1.0*value[11]*other_value[5] + -1.0*value[13]*other_value[23] + -1.0*value[18]*other_value[8] + 1.0*value[24]*other_value[0] + 1.0*value[17]*other_value[9] + 1.0*value[2]*other_value[15] + 1.0*value[31]*other_value[7] + -1.0*value[27]*other_value[19] + 1.0*value[21]*other_value[6] + 1.0*value[23]*other_value[13] + -1.0*value[22]*other_value[14] + -1.0*value[9]*other_value[17]

@cuda.jit(device=True)
def gmt_func_o25(value, other_value):
    return 0 + -1.0*value[29]*other_value[1] + -1.0*value[20]*other_value[8] + -1.0*value[28]*other_value[16] + -1.0*value[16]*other_value[28] + 1.0*value[27]*other_value[17] + -1.0*value[12]*other_value[22] + -1.0*value[30]*other_value[2] + -1.0*value[7]*other_value[21] + -1.0*value[6]*other_value[31] + 1.0*value[5]*other_value[13] + 1.0*value[22]*other_value[12] + 1.0*value[17]*other_value[27] + -1.0*value[26]*other_value[18] + -1.0*value[18]*other_value[26] + 1.0*value[15]*other_value[3] + 1.0*value[3]*other_value[15] + 1.0*value[2]*other_value[30] + -1.0*value[31]*other_value[6] + 1.0*value[1]*other_value[29] + 1.0*value[19]*other_value[9] + 1.0*value[24]*other_value[10] + -1.0*value[4]*other_value[14] + 1.0*value[21]*other_value[7] + 1.0*value[8]*other_value[20] + -1.0*value[10]*other_value[24] + 1.0*value[11]*other_value[23] + 1.0*value[0]*other_value[25] + 1.0*value[13]*other_value[5] + -1.0*value[9]*other_value[19] + -1.0*value[23]*other_value[11] + -1.0*value[14]*other_value[4] + 1.0*value[25]*other_value[0]

@cuda.jit(device=True)
def gmt_func_o26(value, other_value):
    return 0 + 1.0*value[10]*other_value[8] + 1.0*value[13]*other_value[6] + -1.0*value[4]*other_value[16] + -1.0*value[5]*other_value[31] + -1.0*value[15]*other_value[27] + 1.0*value[6]*other_value[13] + 1.0*value[14]*other_value[28] + -1.0*value[23]*other_value[21] + -1.0*value[31]*other_value[5] + -1.0*value[12]*other_value[29] + -1.0*value[11]*other_value[7] + -1.0*value[21]*other_value[23] + 1.0*value[3]*other_value[17] + 1.0*value[0]*other_value[26] + -1.0*value[2]*other_value[19] + 1.0*value[24]*other_value[20] + -1.0*value[17]*other_value[3] + 1.0*value[29]*other_value[12] + 1.0*value[9]*other_value[30] + 1.0*value[19]*other_value[2] + -1.0*value[22]*other_value[1] + 1.0*value[8]*other_value[10] + -1.0*value[18]*other_value[25] + -1.0*value[30]*other_value[9] + -1.0*value[25]*other_value[18] + 1.0*value[16]*other_value[4] + -1.0*value[7]*other_value[11] + -1.0*value[28]*other_value[14] + 1.0*value[26]*other_value[0] + 1.0*value[1]*other_value[22] + 1.0*value[20]*other_value[24] + 1.0*value[27]*other_value[15]

@cuda.jit(device=True)
def gmt_func_o27(value, other_value):
    return 0 + 1.0*value[13]*other_value[28] + -1.0*value[15]*other_value[26] + -1.0*value[4]*other_value[31] + 1.0*value[14]*other_value[6] + -1.0*value[28]*other_value[13] + -1.0*value[30]*other_value[8] + -1.0*value[5]*other_value[16] + 1.0*value[19]*other_value[24] + 1.0*value[27]*other_value[0] + -1.0*value[22]*other_value[21] + 1.0*value[10]*other_value[9] + -1.0*value[23]*other_value[1] + 1.0*value[9]*other_value[10] + 1.0*value[8]*other_value[30] + 1.0*value[16]*other_value[5] + -1.0*value[21]*other_value[22] + 1.0*value[3]*other_value[18] + 1.0*value[1]*other_value[23] + 1.0*value[24]*other_value[19] + 1.0*value[29]*other_value[11] + -1.0*value[11]*other_value[29] + -1.0*value[31]*other_value[4] + -1.0*value[17]*other_value[25] + 1.0*value[20]*other_value[2] + -1.0*value[18]*other_value[3] + 1.0*value[6]*other_value[14] + -1.0*value[12]*other_value[7] + -1.0*value[25]*other_value[17] + -1.0*value[7]*other_value[12] + 1.0*value[0]*other_value[27] + -1.0*value[2]*other_value[20] + 1.0*value[26]*other_value[15]

@cuda.jit(device=True)
def gmt_func_o28(value, other_value):
    return 0 + 1.0*value[0]*other_value[28] + 1.0*value[14]*other_value[26] + 1.0*value[30]*other_value[7] + 1.0*value[4]*other_value[18] + 1.0*value[3]*other_value[31] + -1.0*value[29]*other_value[10] + -1.0*value[18]*other_value[4] + 1.0*value[17]*other_value[5] + 1.0*value[15]*other_value[6] + 1.0*value[31]*other_value[3] + -1.0*value[26]*other_value[14] + -1.0*value[7]*other_value[30] + 1.0*value[20]*other_value[22] + -1.0*value[12]*other_value[8] + 1.0*value[16]*other_value[25] + 1.0*value[25]*other_value[16] + 1.0*value[1]*other_value[24] + 1.0*value[11]*other_value[9] + -1.0*value[24]*other_value[1] + 1.0*value[9]*other_value[11] + -1.0*value[23]*other_value[19] + 1.0*value[21]*other_value[2] + -1.0*value[19]*other_value[23] + 1.0*value[6]*other_value[15] + -1.0*value[8]*other_value[12] + -1.0*value[2]*other_value[21] + 1.0*value[22]*other_value[20] + 1.0*value[28]*other_value[0] + 1.0*value[27]*other_value[13] + -1.0*value[5]*other_value[17] + -1.0*value[13]*other_value[27] + 1.0*value[10]*other_value[29]

@cuda.jit(device=True)
def gmt_func_o29(value, other_value):
    return 0 + 1.0*value[0]*other_value[29] + 1.0*value[15]*other_value[7] + 1.0*value[1]*other_value[25] + 1.0*value[9]*other_value[13] + -1.0*value[2]*other_value[31] + -1.0*value[24]*other_value[16] + -1.0*value[3]*other_value[21] + -1.0*value[8]*other_value[14] + 1.0*value[23]*other_value[17] + -1.0*value[30]*other_value[6] + -1.0*value[10]*other_value[28] + 1.0*value[28]*other_value[10] + 1.0*value[17]*other_value[23] + 1.0*value[11]*other_value[27] + -1.0*value[14]*other_value[8] + -1.0*value[18]*other_value[22] + 1.0*value[4]*other_value[20] + -1.0*value[5]*other_value[19] + -1.0*value[27]*other_value[11] + 1.0*value[19]*other_value[5] + 1.0*value[21]*other_value[3] + 1.0*value[26]*other_value[12] + -1.0*value[22]*other_value[18] + 1.0*value[6]*other_value[30] + -1.0*value[20]*other_value[4] + 1.0*value[7]*other_value[15] + 1.0*value[29]*other_value[0] + -1.0*value[31]*other_value[2] + -1.0*value[16]*other_value[24] + -1.0*value[25]*other_value[1] + -1.0*value[12]*other_value[26] + 1.0*value[13]*other_value[9]

@cuda.jit(device=True)
def gmt_func_o30(value, other_value):
    return 0 + 1.0*value[2]*other_value[25] + 1.0*value[16]*other_value[21] + 1.0*value[10]*other_value[15] + 1.0*value[0]*other_value[30] + 1.0*value[22]*other_value[5] + 1.0*value[9]*other_value[26] + -1.0*value[23]*other_value[4] + 1.0*value[1]*other_value[31] + -1.0*value[8]*other_value[27] + 1.0*value[24]*other_value[3] + -1.0*value[11]*other_value[14] + 1.0*value[21]*other_value[16] + -1.0*value[25]*other_value[2] + 1.0*value[7]*other_value[28] + 1.0*value[12]*other_value[13] + -1.0*value[20]*other_value[17] + 1.0*value[30]*other_value[0] + -1.0*value[6]*other_value[29] + 1.0*value[13]*other_value[12] + 1.0*value[19]*other_value[18] + 1.0*value[27]*other_value[8] + -1.0*value[5]*other_value[22] + 1.0*value[18]*other_value[19] + -1.0*value[14]*other_value[11] + -1.0*value[28]*other_value[7] + 1.0*value[4]*other_value[23] + 1.0*value[29]*other_value[6] + -1.0*value[17]*other_value[20] + -1.0*value[3]*other_value[24] + 1.0*value[15]*other_value[10] + -1.0*value[26]*other_value[9] + 1.0*value[31]*other_value[1]

@cuda.jit(device=True)
def gmt_func_o31(value, other_value):
    return 0 + -1.0*value[24]*other_value[7] + 1.0*value[1]*other_value[30] + -1.0*value[17]*other_value[14] + 1.0*value[15]*other_value[16] + 1.0*value[18]*other_value[13] + -1.0*value[14]*other_value[17] + 1.0*value[13]*other_value[18] + 1.0*value[19]*other_value[12] + -1.0*value[20]*other_value[11] + 1.0*value[12]*other_value[19] + -1.0*value[11]*other_value[20] + 1.0*value[21]*other_value[10] + 1.0*value[10]*other_value[21] + -1.0*value[22]*other_value[9] + 1.0*value[23]*other_value[8] + 1.0*value[31]*other_value[0] + -1.0*value[9]*other_value[22] + 1.0*value[8]*other_value[23] + -1.0*value[7]*other_value[24] + 1.0*value[25]*other_value[6] + 1.0*value[26]*other_value[5] + 1.0*value[6]*other_value[25] + -1.0*value[27]*other_value[4] + 1.0*value[5]*other_value[26] + 1.0*value[28]*other_value[3] + -1.0*value[4]*other_value[27] + -1.0*value[29]*other_value[2] + 1.0*value[3]*other_value[28] + 1.0*value[30]*other_value[1] + -1.0*value[2]*other_value[29] + 1.0*value[0]*other_value[31] + 1.0*value[16]*other_value[15]

@cuda.jit(device=True)
def gmt_func(value, other_value, output):
    output[0] = gmt_func_o0(value,other_value)
    output[1] = gmt_func_o1(value,other_value)
    output[2] = gmt_func_o2(value,other_value)
    output[3] = gmt_func_o3(value,other_value)
    output[4] = gmt_func_o4(value,other_value)
    output[5] = gmt_func_o5(value,other_value)
    output[6] = gmt_func_o6(value,other_value)
    output[7] = gmt_func_o7(value,other_value)
    output[8] = gmt_func_o8(value,other_value)
    output[9] = gmt_func_o9(value,other_value)
    output[10] = gmt_func_o10(value,other_value)
    output[11] = gmt_func_o11(value,other_value)
    output[12] = gmt_func_o12(value,other_value)
    output[13] = gmt_func_o13(value,other_value)
    output[14] = gmt_func_o14(value,other_value)
    output[15] = gmt_func_o15(value,other_value)
    output[16] = gmt_func_o16(value,other_value)
    output[17] = gmt_func_o17(value,other_value)
    output[18] = gmt_func_o18(value,other_value)
    output[19] = gmt_func_o19(value,other_value)
    output[20] = gmt_func_o20(value,other_value)
    output[21] = gmt_func_o21(value,other_value)
    output[22] = gmt_func_o22(value,other_value)
    output[23] = gmt_func_o23(value,other_value)
    output[24] = gmt_func_o24(value,other_value)
    output[25] = gmt_func_o25(value,other_value)
    output[26] = gmt_func_o26(value,other_value)
    output[27] = gmt_func_o27(value,other_value)
    output[28] = gmt_func_o28(value,other_value)
    output[29] = gmt_func_o29(value,other_value)
    output[30] = gmt_func_o30(value,other_value)
    output[31] = gmt_func_o31(value,other_value)

@cuda.jit(device=True)
def omt_func_o0(value, other_value):
    return 0 + 1.0*value[0]*other_value[0]

@cuda.jit(device=True)
def omt_func_o1(value, other_value):
    return 0 + 1.0*value[0]*other_value[1] + 1.0*value[1]*other_value[0]

@cuda.jit(device=True)
def omt_func_o2(value, other_value):
    return 0 + 1.0*value[0]*other_value[2] + 1.0*value[2]*other_value[0]

@cuda.jit(device=True)
def omt_func_o3(value, other_value):
    return 0 + 1.0*value[3]*other_value[0] + 1.0*value[0]*other_value[3]

@cuda.jit(device=True)
def omt_func_o4(value, other_value):
    return 0 + 1.0*value[4]*other_value[0] + 1.0*value[0]*other_value[4]

@cuda.jit(device=True)
def omt_func_o5(value, other_value):
    return 0 + 1.0*value[5]*other_value[0] + 1.0*value[0]*other_value[5]

@cuda.jit(device=True)
def omt_func_o6(value, other_value):
    return 0 + 1.0*value[1]*other_value[2] + 1.0*value[6]*other_value[0] + -1.0*value[2]*other_value[1] + 1.0*value[0]*other_value[6]

@cuda.jit(device=True)
def omt_func_o7(value, other_value):
    return 0 + 1.0*value[1]*other_value[3] + 1.0*value[0]*other_value[7] + -1.0*value[3]*other_value[1] + 1.0*value[7]*other_value[0]

@cuda.jit(device=True)
def omt_func_o8(value, other_value):
    return 0 + 1.0*value[0]*other_value[8] + 1.0*value[1]*other_value[4] + 1.0*value[8]*other_value[0] + -1.0*value[4]*other_value[1]

@cuda.jit(device=True)
def omt_func_o9(value, other_value):
    return 0 + 1.0*value[9]*other_value[0] + 1.0*value[0]*other_value[9] + 1.0*value[1]*other_value[5] + -1.0*value[5]*other_value[1]

@cuda.jit(device=True)
def omt_func_o10(value, other_value):
    return 0 + -1.0*value[3]*other_value[2] + 1.0*value[2]*other_value[3] + 1.0*value[0]*other_value[10] + 1.0*value[10]*other_value[0]

@cuda.jit(device=True)
def omt_func_o11(value, other_value):
    return 0 + -1.0*value[4]*other_value[2] + 1.0*value[2]*other_value[4] + 1.0*value[0]*other_value[11] + 1.0*value[11]*other_value[0]

@cuda.jit(device=True)
def omt_func_o12(value, other_value):
    return 0 + 1.0*value[2]*other_value[5] + 1.0*value[12]*other_value[0] + 1.0*value[0]*other_value[12] + -1.0*value[5]*other_value[2]

@cuda.jit(device=True)
def omt_func_o13(value, other_value):
    return 0 + 1.0*value[13]*other_value[0] + -1.0*value[4]*other_value[3] + 1.0*value[0]*other_value[13] + 1.0*value[3]*other_value[4]

@cuda.jit(device=True)
def omt_func_o14(value, other_value):
    return 0 + 1.0*value[3]*other_value[5] + -1.0*value[5]*other_value[3] + 1.0*value[14]*other_value[0] + 1.0*value[0]*other_value[14]

@cuda.jit(device=True)
def omt_func_o15(value, other_value):
    return 0 + 1.0*value[4]*other_value[5] + -1.0*value[5]*other_value[4] + 1.0*value[15]*other_value[0] + 1.0*value[0]*other_value[15]

@cuda.jit(device=True)
def omt_func_o16(value, other_value):
    return 0 + 1.0*value[10]*other_value[1] + 1.0*value[16]*other_value[0] + 1.0*value[3]*other_value[6] + 1.0*value[6]*other_value[3] + -1.0*value[7]*other_value[2] + 1.0*value[1]*other_value[10] + 1.0*value[0]*other_value[16] + -1.0*value[2]*other_value[7]

@cuda.jit(device=True)
def omt_func_o17(value, other_value):
    return 0 + 1.0*value[11]*other_value[1] + 1.0*value[17]*other_value[0] + 1.0*value[1]*other_value[11] + -1.0*value[8]*other_value[2] + 1.0*value[4]*other_value[6] + -1.0*value[2]*other_value[8] + 1.0*value[6]*other_value[4] + 1.0*value[0]*other_value[17]

@cuda.jit(device=True)
def omt_func_o18(value, other_value):
    return 0 + 1.0*value[1]*other_value[12] + 1.0*value[6]*other_value[5] + 1.0*value[5]*other_value[6] + 1.0*value[0]*other_value[18] + 1.0*value[18]*other_value[0] + 1.0*value[12]*other_value[1] + -1.0*value[2]*other_value[9] + -1.0*value[9]*other_value[2]

@cuda.jit(device=True)
def omt_func_o19(value, other_value):
    return 0 + 1.0*value[4]*other_value[7] + 1.0*value[1]*other_value[13] + 1.0*value[19]*other_value[0] + 1.0*value[7]*other_value[4] + 1.0*value[0]*other_value[19] + 1.0*value[13]*other_value[1] + -1.0*value[8]*other_value[3] + -1.0*value[3]*other_value[8]

@cuda.jit(device=True)
def omt_func_o20(value, other_value):
    return 0 + 1.0*value[7]*other_value[5] + 1.0*value[0]*other_value[20] + -1.0*value[9]*other_value[3] + 1.0*value[20]*other_value[0] + 1.0*value[14]*other_value[1] + 1.0*value[5]*other_value[7] + -1.0*value[3]*other_value[9] + 1.0*value[1]*other_value[14]

@cuda.jit(device=True)
def omt_func_o21(value, other_value):
    return 0 + 1.0*value[1]*other_value[15] + 1.0*value[15]*other_value[1] + -1.0*value[9]*other_value[4] + 1.0*value[21]*other_value[0] + 1.0*value[0]*other_value[21] + 1.0*value[5]*other_value[8] + -1.0*value[4]*other_value[9] + 1.0*value[8]*other_value[5]

@cuda.jit(device=True)
def omt_func_o22(value, other_value):
    return 0 + 1.0*value[13]*other_value[2] + 1.0*value[0]*other_value[22] + 1.0*value[10]*other_value[4] + -1.0*value[11]*other_value[3] + 1.0*value[2]*other_value[13] + -1.0*value[3]*other_value[11] + 1.0*value[22]*other_value[0] + 1.0*value[4]*other_value[10]

@cuda.jit(device=True)
def omt_func_o23(value, other_value):
    return 0 + 1.0*value[5]*other_value[10] + 1.0*value[2]*other_value[14] + 1.0*value[10]*other_value[5] + 1.0*value[0]*other_value[23] + -1.0*value[12]*other_value[3] + -1.0*value[3]*other_value[12] + 1.0*value[14]*other_value[2] + 1.0*value[23]*other_value[0]

@cuda.jit(device=True)
def omt_func_o24(value, other_value):
    return 0 + -1.0*value[12]*other_value[4] + 1.0*value[0]*other_value[24] + 1.0*value[24]*other_value[0] + 1.0*value[15]*other_value[2] + 1.0*value[2]*other_value[15] + 1.0*value[11]*other_value[5] + 1.0*value[5]*other_value[11] + -1.0*value[4]*other_value[12]

@cuda.jit(device=True)
def omt_func_o25(value, other_value):
    return 0 + 1.0*value[25]*other_value[0] + 1.0*value[3]*other_value[15] + 1.0*value[0]*other_value[25] + -1.0*value[4]*other_value[14] + 1.0*value[5]*other_value[13] + 1.0*value[15]*other_value[3] + -1.0*value[14]*other_value[4] + 1.0*value[13]*other_value[5]

@cuda.jit(device=True)
def omt_func_o26(value, other_value):
    return 0 + -1.0*value[17]*other_value[3] + 1.0*value[10]*other_value[8] + -1.0*value[2]*other_value[19] + -1.0*value[4]*other_value[16] + -1.0*value[11]*other_value[7] + 1.0*value[26]*other_value[0] + 1.0*value[3]*other_value[17] + 1.0*value[1]*other_value[22] + 1.0*value[0]*other_value[26] + -1.0*value[22]*other_value[1] + 1.0*value[8]*other_value[10] + 1.0*value[13]*other_value[6] + 1.0*value[19]*other_value[2] + 1.0*value[16]*other_value[4] + -1.0*value[7]*other_value[11] + 1.0*value[6]*other_value[13]

@cuda.jit(device=True)
def omt_func_o27(value, other_value):
    return 0 + -1.0*value[23]*other_value[1] + 1.0*value[14]*other_value[6] + -1.0*value[12]*other_value[7] + 1.0*value[27]*other_value[0] + -1.0*value[18]*other_value[3] + 1.0*value[20]*other_value[2] + 1.0*value[16]*other_value[5] + 1.0*value[10]*other_value[9] + 1.0*value[6]*other_value[14] + 1.0*value[0]*other_value[27] + 1.0*value[1]*other_value[23] + 1.0*value[9]*other_value[10] + -1.0*value[2]*other_value[20] + -1.0*value[5]*other_value[16] + 1.0*value[3]*other_value[18] + -1.0*value[7]*other_value[12]

@cuda.jit(device=True)
def omt_func_o28(value, other_value):
    return 0 + 1.0*value[15]*other_value[6] + -1.0*value[5]*other_value[17] + -1.0*value[24]*other_value[1] + 1.0*value[1]*other_value[24] + 1.0*value[0]*other_value[28] + 1.0*value[4]*other_value[18] + -1.0*value[18]*other_value[4] + 1.0*value[6]*other_value[15] + 1.0*value[21]*other_value[2] + 1.0*value[17]*other_value[5] + 1.0*value[9]*other_value[11] + 1.0*value[11]*other_value[9] + -1.0*value[8]*other_value[12] + -1.0*value[2]*other_value[21] + 1.0*value[28]*other_value[0] + -1.0*value[12]*other_value[8]

@cuda.jit(device=True)
def omt_func_o29(value, other_value):
    return 0 + 1.0*value[19]*other_value[5] + -1.0*value[3]*other_value[21] + -1.0*value[20]*other_value[4] + 1.0*value[0]*other_value[29] + 1.0*value[29]*other_value[0] + 1.0*value[21]*other_value[3] + 1.0*value[15]*other_value[7] + 1.0*value[9]*other_value[13] + 1.0*value[13]*other_value[9] + -1.0*value[5]*other_value[19] + -1.0*value[8]*other_value[14] + 1.0*value[1]*other_value[25] + -1.0*value[14]*other_value[8] + -1.0*value[25]*other_value[1] + 1.0*value[7]*other_value[15] + 1.0*value[4]*other_value[20]

@cuda.jit(device=True)
def omt_func_o30(value, other_value):
    return 0 + 1.0*value[22]*other_value[5] + -1.0*value[25]*other_value[2] + -1.0*value[23]*other_value[4] + -1.0*value[3]*other_value[24] + 1.0*value[24]*other_value[3] + 1.0*value[30]*other_value[0] + 1.0*value[2]*other_value[25] + 1.0*value[10]*other_value[15] + 1.0*value[12]*other_value[13] + 1.0*value[4]*other_value[23] + -1.0*value[11]*other_value[14] + -1.0*value[14]*other_value[11] + -1.0*value[5]*other_value[22] + 1.0*value[0]*other_value[30] + 1.0*value[13]*other_value[12] + 1.0*value[15]*other_value[10]

@cuda.jit(device=True)
def omt_func_o31(value, other_value):
    return 0 + -1.0*value[24]*other_value[7] + 1.0*value[30]*other_value[1] + -1.0*value[7]*other_value[24] + 1.0*value[12]*other_value[19] + 1.0*value[18]*other_value[13] + 1.0*value[26]*other_value[5] + 1.0*value[8]*other_value[23] + -1.0*value[27]*other_value[4] + -1.0*value[11]*other_value[20] + 1.0*value[28]*other_value[3] + -1.0*value[29]*other_value[2] + -1.0*value[9]*other_value[22] + 1.0*value[25]*other_value[6] + 1.0*value[13]*other_value[18] + 1.0*value[6]*other_value[25] + 1.0*value[23]*other_value[8] + -1.0*value[17]*other_value[14] + 1.0*value[3]*other_value[28] + 1.0*value[10]*other_value[21] + 1.0*value[19]*other_value[12] + -1.0*value[4]*other_value[27] + 1.0*value[0]*other_value[31] + -1.0*value[2]*other_value[29] + 1.0*value[16]*other_value[15] + 1.0*value[15]*other_value[16] + 1.0*value[21]*other_value[10] + 1.0*value[5]*other_value[26] + -1.0*value[14]*other_value[17] + -1.0*value[22]*other_value[9] + 1.0*value[1]*other_value[30] + -1.0*value[20]*other_value[11] + 1.0*value[31]*other_value[0]

@cuda.jit(device=True)
def omt_func(value, other_value, output):
    output[0] = omt_func_o0(value,other_value)
    output[1] = omt_func_o1(value,other_value)
    output[2] = omt_func_o2(value,other_value)
    output[3] = omt_func_o3(value,other_value)
    output[4] = omt_func_o4(value,other_value)
    output[5] = omt_func_o5(value,other_value)
    output[6] = omt_func_o6(value,other_value)
    output[7] = omt_func_o7(value,other_value)
    output[8] = omt_func_o8(value,other_value)
    output[9] = omt_func_o9(value,other_value)
    output[10] = omt_func_o10(value,other_value)
    output[11] = omt_func_o11(value,other_value)
    output[12] = omt_func_o12(value,other_value)
    output[13] = omt_func_o13(value,other_value)
    output[14] = omt_func_o14(value,other_value)
    output[15] = omt_func_o15(value,other_value)
    output[16] = omt_func_o16(value,other_value)
    output[17] = omt_func_o17(value,other_value)
    output[18] = omt_func_o18(value,other_value)
    output[19] = omt_func_o19(value,other_value)
    output[20] = omt_func_o20(value,other_value)
    output[21] = omt_func_o21(value,other_value)
    output[22] = omt_func_o22(value,other_value)
    output[23] = omt_func_o23(value,other_value)
    output[24] = omt_func_o24(value,other_value)
    output[25] = omt_func_o25(value,other_value)
    output[26] = omt_func_o26(value,other_value)
    output[27] = omt_func_o27(value,other_value)
    output[28] = omt_func_o28(value,other_value)
    output[29] = omt_func_o29(value,other_value)
    output[30] = omt_func_o30(value,other_value)
    output[31] = omt_func_o31(value,other_value)

@cuda.jit(device=True)
def imt_func_o0(value, other_value):
    return 0 + 1.0*value[1]*other_value[1] + -1.0*value[5]*other_value[5] + -1.0*value[6]*other_value[6] + -1.0*value[7]*other_value[7] + -1.0*value[8]*other_value[8] + 1.0*value[9]*other_value[9] + -1.0*value[10]*other_value[10] + -1.0*value[11]*other_value[11] + 1.0*value[12]*other_value[12] + -1.0*value[13]*other_value[13] + 1.0*value[14]*other_value[14] + 1.0*value[15]*other_value[15] + -1.0*value[16]*other_value[16] + 1.0*value[4]*other_value[4] + -1.0*value[17]*other_value[17] + -1.0*value[19]*other_value[19] + 1.0*value[20]*other_value[20] + 1.0*value[21]*other_value[21] + -1.0*value[22]*other_value[22] + 1.0*value[23]*other_value[23] + 1.0*value[24]*other_value[24] + 1.0*value[25]*other_value[25] + 1.0*value[26]*other_value[26] + -1.0*value[27]*other_value[27] + -1.0*value[28]*other_value[28] + -1.0*value[29]*other_value[29] + -1.0*value[30]*other_value[30] + 1.0*value[18]*other_value[18] + 1.0*value[3]*other_value[3] + -1.0*value[31]*other_value[31] + 1.0*value[2]*other_value[2]

@cuda.jit(device=True)
def imt_func_o1(value, other_value):
    return 0 + -1.0*value[17]*other_value[11] + 1.0*value[5]*other_value[9] + 1.0*value[22]*other_value[26] + -1.0*value[26]*other_value[22] + 1.0*value[6]*other_value[2] + 1.0*value[29]*other_value[25] + 1.0*value[7]*other_value[3] + 1.0*value[21]*other_value[15] + 1.0*value[8]*other_value[4] + 1.0*value[20]*other_value[14] + -1.0*value[9]*other_value[5] + -1.0*value[19]*other_value[13] + -1.0*value[10]*other_value[16] + 1.0*value[28]*other_value[24] + -1.0*value[11]*other_value[17] + 1.0*value[12]*other_value[18] + 1.0*value[18]*other_value[12] + -1.0*value[13]*other_value[19] + -1.0*value[2]*other_value[6] + 1.0*value[14]*other_value[20] + 1.0*value[27]*other_value[23] + -1.0*value[31]*other_value[30] + 1.0*value[15]*other_value[21] + -1.0*value[30]*other_value[31] + -1.0*value[23]*other_value[27] + -1.0*value[16]*other_value[10] + -1.0*value[3]*other_value[7] + -1.0*value[24]*other_value[28] + -1.0*value[25]*other_value[29] + -1.0*value[4]*other_value[8]

@cuda.jit(device=True)
def imt_func_o2(value, other_value):
    return 0 + 1.0*value[10]*other_value[3] + 1.0*value[24]*other_value[15] + 1.0*value[7]*other_value[16] + 1.0*value[21]*other_value[28] + 1.0*value[8]*other_value[17] + 1.0*value[16]*other_value[7] + -1.0*value[9]*other_value[18] + -1.0*value[19]*other_value[26] + -1.0*value[28]*other_value[21] + -1.0*value[12]*other_value[5] + -1.0*value[25]*other_value[30] + -1.0*value[13]*other_value[22] + -1.0*value[18]*other_value[9] + -1.0*value[27]*other_value[20] + 1.0*value[14]*other_value[23] + -1.0*value[3]*other_value[10] + 1.0*value[31]*other_value[29] + 1.0*value[15]*other_value[24] + 1.0*value[17]*other_value[8] + 1.0*value[11]*other_value[4] + 1.0*value[29]*other_value[31] + 1.0*value[20]*other_value[27] + 1.0*value[1]*other_value[6] + 1.0*value[5]*other_value[12] + -1.0*value[6]*other_value[1] + 1.0*value[23]*other_value[14] + 1.0*value[26]*other_value[19] + -1.0*value[22]*other_value[13] + -1.0*value[4]*other_value[11] + 1.0*value[30]*other_value[25]

@cuda.jit(device=True)
def imt_func_o3(value, other_value):
    return 0 + 1.0*value[27]*other_value[18] + -1.0*value[18]*other_value[27] + -1.0*value[12]*other_value[23] + -1.0*value[6]*other_value[16] + 1.0*value[19]*other_value[8] + -1.0*value[10]*other_value[2] + -1.0*value[28]*other_value[31] + 1.0*value[25]*other_value[15] + 1.0*value[5]*other_value[14] + 1.0*value[2]*other_value[10] + -1.0*value[9]*other_value[20] + -1.0*value[30]*other_value[24] + -1.0*value[14]*other_value[5] + 1.0*value[17]*other_value[26] + -1.0*value[31]*other_value[28] + -1.0*value[23]*other_value[12] + 1.0*value[15]*other_value[25] + -1.0*value[26]*other_value[17] + -1.0*value[16]*other_value[6] + 1.0*value[13]*other_value[4] + -1.0*value[4]*other_value[13] + 1.0*value[11]*other_value[22] + 1.0*value[8]*other_value[19] + 1.0*value[24]*other_value[30] + -1.0*value[20]*other_value[9] + 1.0*value[1]*other_value[7] + -1.0*value[29]*other_value[21] + -1.0*value[7]*other_value[1] + 1.0*value[21]*other_value[29] + 1.0*value[22]*other_value[11]

@cuda.jit(device=True)
def imt_func_o4(value, other_value):
    return 0 + -1.0*value[8]*other_value[1] + -1.0*value[12]*other_value[24] + -1.0*value[18]*other_value[28] + -1.0*value[13]*other_value[3] + -1.0*value[21]*other_value[9] + -1.0*value[24]*other_value[12] + -1.0*value[7]*other_value[19] + -1.0*value[14]*other_value[25] + -1.0*value[22]*other_value[10] + 1.0*value[5]*other_value[15] + -1.0*value[15]*other_value[5] + 1.0*value[29]*other_value[20] + -1.0*value[17]*other_value[6] + -1.0*value[16]*other_value[26] + 1.0*value[27]*other_value[31] + -1.0*value[6]*other_value[17] + 1.0*value[3]*other_value[13] + 1.0*value[2]*other_value[11] + 1.0*value[26]*other_value[16] + -1.0*value[20]*other_value[29] + -1.0*value[25]*other_value[14] + -1.0*value[10]*other_value[22] + -1.0*value[11]*other_value[2] + 1.0*value[31]*other_value[27] + 1.0*value[30]*other_value[23] + -1.0*value[23]*other_value[30] + 1.0*value[1]*other_value[8] + -1.0*value[19]*other_value[7] + -1.0*value[9]*other_value[21] + 1.0*value[28]*other_value[18]

@cuda.jit(device=True)
def imt_func_o5(value, other_value):
    return 0 + 1.0*value[4]*other_value[15] + 1.0*value[1]*other_value[9] + -1.0*value[16]*other_value[27] + 1.0*value[31]*other_value[26] + -1.0*value[19]*other_value[29] + -1.0*value[15]*other_value[4] + -1.0*value[24]*other_value[11] + 1.0*value[29]*other_value[19] + -1.0*value[17]*other_value[28] + -1.0*value[10]*other_value[23] + -1.0*value[9]*other_value[1] + -1.0*value[6]*other_value[18] + -1.0*value[12]*other_value[2] + 1.0*value[3]*other_value[14] + -1.0*value[14]*other_value[3] + -1.0*value[23]*other_value[10] + 1.0*value[27]*other_value[16] + 1.0*value[30]*other_value[22] + -1.0*value[11]*other_value[24] + -1.0*value[21]*other_value[8] + -1.0*value[8]*other_value[21] + 1.0*value[28]*other_value[17] + -1.0*value[20]*other_value[7] + 1.0*value[26]*other_value[31] + -1.0*value[13]*other_value[25] + -1.0*value[22]*other_value[30] + -1.0*value[18]*other_value[6] + -1.0*value[7]*other_value[20] + 1.0*value[2]*other_value[12] + -1.0*value[25]*other_value[13]

@cuda.jit(device=True)
def imt_func_o6(value, other_value):
    return 0 + -1.0*value[26]*other_value[13] + 1.0*value[25]*other_value[31] + 1.0*value[17]*other_value[4] + 1.0*value[27]*other_value[14] + -1.0*value[18]*other_value[5] + -1.0*value[13]*other_value[26] + 1.0*value[3]*other_value[16] + 1.0*value[31]*other_value[25] + 1.0*value[15]*other_value[28] + 1.0*value[16]*other_value[3] + 1.0*value[28]*other_value[15] + 1.0*value[14]*other_value[27] + 1.0*value[4]*other_value[17] + -1.0*value[5]*other_value[18]

@cuda.jit(device=True)
def imt_func_o7(value, other_value):
    return 0 + -1.0*value[16]*other_value[2] + -1.0*value[20]*other_value[5] + 1.0*value[19]*other_value[4] + 1.0*value[11]*other_value[26] + -1.0*value[12]*other_value[27] + 1.0*value[29]*other_value[15] + -1.0*value[5]*other_value[20] + -1.0*value[2]*other_value[16] + 1.0*value[4]*other_value[19] + -1.0*value[24]*other_value[31] + -1.0*value[31]*other_value[24] + 1.0*value[15]*other_value[29] + 1.0*value[26]*other_value[11] + -1.0*value[27]*other_value[12]

@cuda.jit(device=True)
def imt_func_o8(value, other_value):
    return 0 + -1.0*value[5]*other_value[21] + 1.0*value[31]*other_value[23] + -1.0*value[21]*other_value[5] + -1.0*value[2]*other_value[17] + -1.0*value[26]*other_value[10] + -1.0*value[3]*other_value[19] + 1.0*value[23]*other_value[31] + -1.0*value[29]*other_value[14] + -1.0*value[17]*other_value[2] + -1.0*value[28]*other_value[12] + -1.0*value[12]*other_value[28] + -1.0*value[19]*other_value[3] + -1.0*value[14]*other_value[29] + -1.0*value[10]*other_value[26]

@cuda.jit(device=True)
def imt_func_o9(value, other_value):
    return 0 + -1.0*value[2]*other_value[18] + -1.0*value[11]*other_value[28] + -1.0*value[21]*other_value[4] + 1.0*value[31]*other_value[22] + 1.0*value[22]*other_value[31] + -1.0*value[13]*other_value[29] + -1.0*value[29]*other_value[13] + -1.0*value[28]*other_value[11] + -1.0*value[10]*other_value[27] + -1.0*value[20]*other_value[3] + -1.0*value[27]*other_value[10] + -1.0*value[4]*other_value[21] + -1.0*value[18]*other_value[2] + -1.0*value[3]*other_value[20]

@cuda.jit(device=True)
def imt_func_o10(value, other_value):
    return 0 + 1.0*value[30]*other_value[15] + 1.0*value[31]*other_value[21] + -1.0*value[23]*other_value[5] + -1.0*value[5]*other_value[23] + 1.0*value[22]*other_value[4] + 1.0*value[4]*other_value[22] + 1.0*value[1]*other_value[16] + 1.0*value[21]*other_value[31] + 1.0*value[15]*other_value[30] + -1.0*value[8]*other_value[26] + 1.0*value[27]*other_value[9] + 1.0*value[16]*other_value[1] + -1.0*value[26]*other_value[8] + 1.0*value[9]*other_value[27]

@cuda.jit(device=True)
def imt_func_o11(value, other_value):
    return 0 + -1.0*value[14]*other_value[30] + -1.0*value[3]*other_value[22] + 1.0*value[1]*other_value[17] + -1.0*value[24]*other_value[5] + 1.0*value[28]*other_value[9] + 1.0*value[17]*other_value[1] + 1.0*value[9]*other_value[28] + -1.0*value[31]*other_value[20] + -1.0*value[20]*other_value[31] + 1.0*value[7]*other_value[26] + -1.0*value[30]*other_value[14] + -1.0*value[5]*other_value[24] + 1.0*value[26]*other_value[7] + -1.0*value[22]*other_value[3]

@cuda.jit(device=True)
def imt_func_o12(value, other_value):
    return 0 + 1.0*value[18]*other_value[1] + -1.0*value[19]*other_value[31] + 1.0*value[28]*other_value[8] + 1.0*value[8]*other_value[28] + -1.0*value[31]*other_value[19] + 1.0*value[27]*other_value[7] + -1.0*value[3]*other_value[23] + -1.0*value[23]*other_value[3] + 1.0*value[7]*other_value[27] + -1.0*value[4]*other_value[24] + -1.0*value[30]*other_value[13] + -1.0*value[24]*other_value[4] + -1.0*value[13]*other_value[30] + 1.0*value[1]*other_value[18]

@cuda.jit(device=True)
def imt_func_o13(value, other_value):
    return 0 + 1.0*value[30]*other_value[12] + 1.0*value[29]*other_value[9] + 1.0*value[31]*other_value[18] + -1.0*value[26]*other_value[6] + 1.0*value[19]*other_value[1] + 1.0*value[18]*other_value[31] + -1.0*value[6]*other_value[26] + -1.0*value[5]*other_value[25] + 1.0*value[2]*other_value[22] + 1.0*value[22]*other_value[2] + 1.0*value[1]*other_value[19] + 1.0*value[12]*other_value[30] + -1.0*value[25]*other_value[5] + 1.0*value[9]*other_value[29]

@cuda.jit(device=True)
def imt_func_o14(value, other_value):
    return 0 + -1.0*value[27]*other_value[6] + 1.0*value[17]*other_value[31] + 1.0*value[29]*other_value[8] + 1.0*value[2]*other_value[23] + 1.0*value[20]*other_value[1] + -1.0*value[6]*other_value[27] + 1.0*value[11]*other_value[30] + 1.0*value[30]*other_value[11] + 1.0*value[31]*other_value[17] + 1.0*value[8]*other_value[29] + -1.0*value[4]*other_value[25] + -1.0*value[25]*other_value[4] + 1.0*value[1]*other_value[20] + 1.0*value[23]*other_value[2]

@cuda.jit(device=True)
def imt_func_o15(value, other_value):
    return 0 + -1.0*value[28]*other_value[6] + -1.0*value[30]*other_value[10] + -1.0*value[6]*other_value[28] + -1.0*value[10]*other_value[30] + -1.0*value[7]*other_value[29] + -1.0*value[29]*other_value[7] + -1.0*value[16]*other_value[31] + 1.0*value[2]*other_value[24] + 1.0*value[3]*other_value[25] + 1.0*value[25]*other_value[3] + -1.0*value[31]*other_value[16] + 1.0*value[21]*other_value[1] + 1.0*value[24]*other_value[2] + 1.0*value[1]*other_value[21]

@cuda.jit(device=True)
def imt_func_o16(value, other_value):
    return 0 + 1.0*value[26]*other_value[4] + -1.0*value[27]*other_value[5] + 1.0*value[5]*other_value[27] + 1.0*value[15]*other_value[31] + 1.0*value[31]*other_value[15] + -1.0*value[4]*other_value[26]

@cuda.jit(device=True)
def imt_func_o17(value, other_value):
    return 0 + 1.0*value[3]*other_value[26] + -1.0*value[31]*other_value[14] + 1.0*value[5]*other_value[28] + -1.0*value[26]*other_value[3] + -1.0*value[28]*other_value[5] + -1.0*value[14]*other_value[31]

@cuda.jit(device=True)
def imt_func_o18(value, other_value):
    return 0 + -1.0*value[31]*other_value[13] + -1.0*value[27]*other_value[3] + 1.0*value[4]*other_value[28] + 1.0*value[3]*other_value[27] + -1.0*value[28]*other_value[4] + -1.0*value[13]*other_value[31]

@cuda.jit(device=True)
def imt_func_o19(value, other_value):
    return 0 + -1.0*value[29]*other_value[5] + 1.0*value[31]*other_value[12] + -1.0*value[2]*other_value[26] + 1.0*value[5]*other_value[29] + 1.0*value[26]*other_value[2] + 1.0*value[12]*other_value[31]

@cuda.jit(device=True)
def imt_func_o20(value, other_value):
    return 0 + -1.0*value[29]*other_value[4] + 1.0*value[4]*other_value[29] + 1.0*value[27]*other_value[2] + 1.0*value[31]*other_value[11] + -1.0*value[2]*other_value[27] + 1.0*value[11]*other_value[31]

@cuda.jit(device=True)
def imt_func_o21(value, other_value):
    return 0 + 1.0*value[28]*other_value[2] + -1.0*value[10]*other_value[31] + -1.0*value[2]*other_value[28] + -1.0*value[3]*other_value[29] + -1.0*value[31]*other_value[10] + 1.0*value[29]*other_value[3]

@cuda.jit(device=True)
def imt_func_o22(value, other_value):
    return 0 + 1.0*value[5]*other_value[30] + -1.0*value[30]*other_value[5] + -1.0*value[31]*other_value[9] + -1.0*value[26]*other_value[1] + 1.0*value[1]*other_value[26] + -1.0*value[9]*other_value[31]

@cuda.jit(device=True)
def imt_func_o23(value, other_value):
    return 0 + -1.0*value[31]*other_value[8] + 1.0*value[4]*other_value[30] + -1.0*value[27]*other_value[1] + -1.0*value[30]*other_value[4] + -1.0*value[8]*other_value[31] + 1.0*value[1]*other_value[27]

@cuda.jit(device=True)
def imt_func_o24(value, other_value):
    return 0 + 1.0*value[31]*other_value[7] + 1.0*value[7]*other_value[31] + 1.0*value[1]*other_value[28] + -1.0*value[3]*other_value[30] + -1.0*value[28]*other_value[1] + 1.0*value[30]*other_value[3]

@cuda.jit(device=True)
def imt_func_o25(value, other_value):
    return 0 + 1.0*value[2]*other_value[30] + 1.0*value[1]*other_value[29] + -1.0*value[30]*other_value[2] + -1.0*value[6]*other_value[31] + -1.0*value[29]*other_value[1] + -1.0*value[31]*other_value[6]

@cuda.jit(device=True)
def imt_func_o26(value, other_value):
    return 0 + -1.0*value[31]*other_value[5] + -1.0*value[5]*other_value[31]

@cuda.jit(device=True)
def imt_func_o27(value, other_value):
    return 0 + -1.0*value[31]*other_value[4] + -1.0*value[4]*other_value[31]

@cuda.jit(device=True)
def imt_func_o28(value, other_value):
    return 0 + 1.0*value[31]*other_value[3] + 1.0*value[3]*other_value[31]

@cuda.jit(device=True)
def imt_func_o29(value, other_value):
    return 0 + -1.0*value[31]*other_value[2] + -1.0*value[2]*other_value[31]

@cuda.jit(device=True)
def imt_func_o30(value, other_value):
    return 0 + 1.0*value[1]*other_value[31] + 1.0*value[31]*other_value[1]

@cuda.jit(device=True)
def imt_func_o31(value, other_value):
    return 0

@cuda.jit(device=True)
def imt_func(value, other_value, output):
    output[0] = imt_func_o0(value,other_value)
    output[1] = imt_func_o1(value,other_value)
    output[2] = imt_func_o2(value,other_value)
    output[3] = imt_func_o3(value,other_value)
    output[4] = imt_func_o4(value,other_value)
    output[5] = imt_func_o5(value,other_value)
    output[6] = imt_func_o6(value,other_value)
    output[7] = imt_func_o7(value,other_value)
    output[8] = imt_func_o8(value,other_value)
    output[9] = imt_func_o9(value,other_value)
    output[10] = imt_func_o10(value,other_value)
    output[11] = imt_func_o11(value,other_value)
    output[12] = imt_func_o12(value,other_value)
    output[13] = imt_func_o13(value,other_value)
    output[14] = imt_func_o14(value,other_value)
    output[15] = imt_func_o15(value,other_value)
    output[16] = imt_func_o16(value,other_value)
    output[17] = imt_func_o17(value,other_value)
    output[18] = imt_func_o18(value,other_value)
    output[19] = imt_func_o19(value,other_value)
    output[20] = imt_func_o20(value,other_value)
    output[21] = imt_func_o21(value,other_value)
    output[22] = imt_func_o22(value,other_value)
    output[23] = imt_func_o23(value,other_value)
    output[24] = imt_func_o24(value,other_value)
    output[25] = imt_func_o25(value,other_value)
    output[26] = imt_func_o26(value,other_value)
    output[27] = imt_func_o27(value,other_value)
    output[28] = imt_func_o28(value,other_value)
    output[29] = imt_func_o29(value,other_value)
    output[30] = imt_func_o30(value,other_value)
    output[31] = imt_func_o31(value,other_value)

